/*
 * Copyright (c) 2002 - 2005 Regents of The University of Michigan.
 * All Rights Reserved.  See COPYRIGHT.
 */
#define _WIN32_WINNT  0x0500
#define WINVER 0x0500

#include <windows.h>
#include <httpfilt.h>
#include <stdio.h>
#include <malloc.h>
#include <openssl\ssl.h>
#include <openssl\err.h>
#include <libsnet\snet.h>
#include "CosignMsg.h"
#include "CosignFilterMain.h"
#include "CosignStructs.h"
#include "config.h"
#include "network.h"
#include "sparse.h"
#include "mkcookie.h"
#include "version.h"

#define CSCONFIGKEY	"SOFTWARE\\University of Michigan\\ITCS\\Cosign"
#define CSCONFIGFILE "ConfigFile"

#define CSINITMSG   "Starting Cosign Filter version " VERSION "\n"

COSIGNSETTINGS      g_cs;
struct connlist*    g_cl;
CRITICAL_SECTION	g_csConn;
CRITICAL_SECTION	g_csConnPool;
SSL_CTX*            g_ctx;
int                 g_numConnections;
struct connpool*    g_cp;
SERVICELIST	        g_sl;

int
CookieValid( char*, struct connlist **, SERVICECONFIGDATA*, COSIGNSETTINGS*, struct sinfo *, char* , SSL_CTX* );

/*
 * FindCookie()
 *
 * Searches for the cookie value labeled szServiceName.  If there
 * is not a match or the value is not set, the function returns 0.
 * If the buffer szCookie is too small, returns -1.
 *
 * If everything goes smoothly the cookie will be copied into the
 * szCookie buffer supplied by the caller.  The size of szCookie
 * is specified by dwCookieSize.
 */
	int
FindCookie( char* szServiceName, char* szCookieData,
		   char* szCookie, DWORD dwCookieSize ) 
{

	char*	pServiceName = NULL;
	char*	pCookie = NULL;
	char*	pSeparator = NULL;

	if( strlen( szServiceName ) < 1 ||
		strlen( szCookieData ) < 1 ) {
		return( 0 );
	}
	if ( (pServiceName = strstr( szCookieData, szServiceName )) == NULL )  {
		return( 0 );
	}
	if ( (pCookie = strchr( pServiceName, '=' )) == NULL ) {
		return( 0 );
	}
	*pCookie = '\0';
	pCookie++;
	
	//remove any leading spaces
	while( *pCookie == ' ' ) { pCookie++; }

	//Make sure we're left with something
	if ( pCookie == NULL ) {
		return( 0 );
	}
	pSeparator = strchr( pCookie, ';' );
	if ( pSeparator ) {
		*pSeparator = '\0';
	}
	if( strlen( pCookie ) + 1 > dwCookieSize ) {
		///Oh no!  The buffer is too small yet again!
		CosignLog( CS_ERROR, "Buffer for parsing service cookie too small.\n" );
		return( -1 );
	}
	strcpy( szCookie, pCookie );

	return( (int)strlen( szCookie ) );	
}

/*
 * SetCookieAndRedirect()
 *
 * PHTTP_FILTER_CONTEXT pfc - Information regarding this browser session.  Needed
 * for support functions.
 *
 * PHTTP_FILTER_PREPROC_HEADERS pfph - Information regarding this phase of the
 * IIS request-response cycle.  Needed to get end-user/browser specific information.
 *
 * SERVICECONFIGDATA* scd - Service data associated with the cosign-protected service
 * the user is trying to access.
 *
 * returns DWORD - 0, everything is fine.  -1, something is not so fine.
 *
 * Generates a random cookie value identifying the user.  Also sends a redirect
 * message to the user's web browser.  User is redirected to the central cosign
 * login web page.
 *
 */
DWORD
SetCookieAndRedirect( PHTTP_FILTER_CONTEXT pfc, PHTTP_FILTER_PREPROC_HEADERS pfph, SERVICECONFIGDATA* scd )
{
	char	szCookie[ 128 ];
	char	szHeaders[ 2048 ];
	char	szLocation[ 2048 ];
	char	szUrlBuff[ DEFAULTBUFFERSIZE ];
    char	szUrlBuff2[ DEFAULTBUFFERSIZE ];
	DWORD	dwUrlBuffSize = DEFAULTBUFFERSIZE;
	char	szHost[ 1024 ];
    // "http" or "https" plus '\0' <= 6 characters
	char	szProtocol[ 6 ];
	DWORD	dwHostSize = 1024;
    char    szMethod[ DEFAULTBUFFERSIZE ];
    char*   szUrl;
    DWORD   dwSize = DEFAULTBUFFERSIZE;

    if ( !pfc->GetServerVariable( pfc, "REQUEST_METHOD", szMethod, &dwSize ) ) {
        CosignLog( GetLastError(), "GetServerVariable(\"REQUEST_METHOD\") failed.\n" );
        return( -1 );
    }
    CSDEBUG1( CS_INFO, "REQUEST_METHOD = %s.\n", szMethod );
    
    if ( _stricmp( szMethod, "POST" ) == 0 ) {
        CSDEBUG0( CS_INFO, "Redirecting from POST, user may lose some data!\n");
        if ( _snprintf( szHeaders, sizeof( szHeaders ), 
	        "Location: %s\r\n\r\n", g_cs.cs_postErrorRedirectURL ) > sizeof( szHeaders ) ) {
	            CosignLog( CS_ERROR, "Buffer for redirect headers too small.\n" );
	            return( -1 );
            }
        pfc->ServerSupportFunction( pfc, SF_REQ_SEND_RESPONSE_HEADER, "302 Redirect.", (DWORD)szHeaders, 0 );
        CSDEBUG0( CS_INFO, "Added the POST response headers\n" );   
        return( 0 );
    } 
    // If not POST, assume GET, though, note that IIS 5.1 doesn't always populate REQUEST_METHOD.

	//Generate a random cookie to identify the user
	if ( mkcookie( sizeof( szCookie ), szCookie ) != 0 ) {
		CosignLog( CS_ERROR, "mkcookie() failed!  Could not generate a random service cookie.\n" );
		return( -1 );
	}

    if ( scd->scd_siteEntry != NULL ) {
        szUrl = scd->scd_siteEntry;
    } else if ( pfph->GetHeader( pfc, "url", szUrlBuff, &dwUrlBuffSize ) ) {
        
        if ( !pfc->GetServerVariable( pfc, "HTTP_HOST", szHost, &dwHostSize ) ) {
		    CosignLog( GetLastError(), "GetHeader(\"HTTP_HOST\") failed.\n" );
		    return( -1 );
    	}
	    strcpy( szProtocol, "http" );
	    if ( g_cs.cs_redirectToHttps ) {
		    CSDEBUG1( CS_INFO, "httpsonly is set to %d.  Redirecting to https.", g_cs.cs_redirectToHttps);
		    strcpy( szProtocol, "https" );
	    }
	    if ( pfc->fIsSecurePort ) {
		    CSDEBUG1( CS_INFO, "fIsSecurePort is %d.  Redirecting to https.", pfc->fIsSecurePort );
		    strcpy( szProtocol, "https" );
	    }
        if ( _snprintf( szUrlBuff2, sizeof( szUrlBuff2 ), "%s://%s%s", szProtocol, szHost, szUrlBuff ) > sizeof( szUrlBuff2 ) ) {
            CosignLog( CS_ERROR, "Buffer for redirect URL too small.\n" );
            return( -1 );
        }
        szUrl = szUrlBuff2;
    } else {
	    CosignLog( GetLastError(), "GetHeader(\"url\") failed.\n" );
	    return( -1 );
    }

    if ( scd->strFactors ) {
        if ( _snprintf( szLocation, sizeof( szLocation ), "%sfactors=%s&%s=%s&%s", g_cs.cs_redirectURL, scd->strFactors, scd->scd_name, szCookie, szUrl ) > sizeof( szLocation ) ) {
			CosignLog( CS_ERROR, "Buffer for location URL (with factors) too small.\n" );
			return( -1 );
		}
	} else {
		if ( _snprintf( szLocation, sizeof( szLocation ), "%s%s=%s&%s", g_cs.cs_redirectURL, scd->scd_name,
			szCookie, szUrl ) > sizeof( szLocation ) ) {
			CosignLog( CS_ERROR, "Buffer for location URL too small.\n" );
			return( -1 );
		}
	}

    CSDEBUG1( CS_INFO, "Here's the URL: %s\n", szLocation );
    //Send the cookie to the client and redirect them to the central
	//sign-in server.

	if ( _snprintf( szHeaders, sizeof( szHeaders ), "Set-Cookie: %s=%s;"
		"path=/%s"
		"Location: %s\r\n\r\n", scd->scd_name, szCookie, g_cs.cs_cookieOptions, szLocation ) > sizeof( szHeaders ) ) {
	    CosignLog( CS_ERROR, "Buffer for redirect headers too small.\n" );
	    return( -1 );
    }


	pfc->ServerSupportFunction( pfc, SF_REQ_SEND_RESPONSE_HEADER, "302 Redirect.", (DWORD)szHeaders, 0 );

	CSDEBUG1( CS_INFO, "Added the response headers:\n%s\n", szHeaders );
	return( 0 );
}

/*
 * OnPreprocHeaders
 *
 * in variables - IIS-provided state information.  Necessary to keep track
 * of whom filter is responding to.
 *
 * returns DWORD - Lets calling function know if everything went smoothly.  These
 * are standard responses a filter is allowed to give.  In this case, it's basically
 * SF_STATUS_REQ_NEXT_NOTIFICATION if the URL is not cosign-protected or the user
 * has been successfully identifed.
 *
 * When filter receives the SF_NOTIFY_PREPROC_HEADERS notification, this is the
 * main function for handling the 'cosign-protecting' duties.  SF_STATUS_REQ_FINISHED 
 * is returned when either the user was redirected or there was an error in 
 * processing.  Major errors are put into the log file.
 */
DWORD
OnPreprocHeaders( PHTTP_FILTER_CONTEXT pfc, PHTTP_FILTER_PREPROC_HEADERS pfph )
{

	char	szBuffer[ DEFAULTBUFFERSIZE ] = "";
	char	szInstanceID[ DEFAULTBUFFERSIZE ] = "";
	char	szCookie[ DEFAULTCOOKIESIZE + 1 ] = "";
	char	szIpAddr[ DEFAULTCOOKIESIZE ]; 
	char	szUrl[ DEFAULTBUFFERSIZE ];
	char	krbpath [ MAX_PATH + 5]; // +5 is for the additional "FILE:" prefix
	char	proxypath [ MAX_PATH ];
    char*	token;
	DWORD	dwUrlSize = DEFAULTBUFFERSIZE;
	DWORD	dwSize = DEFAULTBUFFERSIZE;
	DWORD	dwInstanceID;
	int		protectedStatus;
	SERVICECONFIGDATA*	scd;
	struct sinfo	si;
	int		cv;
	struct connpool *cp;

    if ( g_cs.cs_runInDefaultMode ) {
        SetLastError( ERROR_SERVER_DISABLED );
		return( SF_STATUS_REQ_ERROR );
	}
	memset((void *)&si, 0, sizeof(si)); // zero out contents of si

	//First determine which web site they're trying to connect to and then
	//give FindCookie() the appropriate service name.
	dwSize = DEFAULTBUFFERSIZE;
	if ( !pfc->GetServerVariable( pfc, "SERVER_NAME", szBuffer, &dwSize ) ) {
		CosignLog( GetLastError(), "OnPreprocHeaders::GetServerVariable( \"SERVER_NAME\" ) failed.\n" );
		SetLastError( ERROR_INTERNAL_ERROR );
		return( SF_STATUS_REQ_ERROR );
	}

	dwSize = DEFAULTBUFFERSIZE;
	if ( !pfc->GetServerVariable( pfc, "INSTANCE_ID", szInstanceID, &dwSize ) ) {
		CosignLog( GetLastError(), "OnPreprocHeaders::GetServerVariable( \"INSTANCE_ID\" ) failed.\n" );
		SetLastError( ERROR_INTERNAL_ERROR );
		return( SF_STATUS_REQ_ERROR );
	}
	dwInstanceID = (DWORD)atol( szInstanceID );

	if ( !pfph->GetHeader( pfc, "url", szUrl, &dwUrlSize ) ) {
		CosignLog( GetLastError(), "OnPreprocHeaders::GetHeader( \"url\" ) failed.\n" );
		SetLastError( ERROR_INTERNAL_ERROR );
		return( SF_STATUS_REQ_ERROR );
	}
	CSDEBUG1( CS_INFO, "url=\"%s\"\n", szUrl );

	if ( !pfc->ServerSupportFunction( pfc, SF_REQ_NORMALIZE_URL, szUrl, 0, 0 ) ) {
		CosignLog( CS_ERROR, "Could not normalize URL: %s\n", szUrl );
		SetLastError( ERROR_INTERNAL_ERROR );
		return( SF_STATUS_REQ_ERROR );
	}
	for ( token = strtok( szUrl , "\\"); token != NULL; token = strtok( NULL, "\\") ) {
        if ( token == szUrl ) continue;
		token[ -1 ] = '/';
	}
	CSDEBUG3( CS_INFO, "g_sl.getProtectedStatus( \"%s\", %d, \"%s\" )\n" , szBuffer, dwInstanceID, szUrl );

	protectedStatus = g_sl.getProtectedStatus( szBuffer, dwInstanceID, szUrl, &scd );
    CSDEBUG1( CS_INFO, "Got the status and it is %d\n", protectedStatus );
	switch ( protectedStatus ) {
		case UNPROTECTED :
			CSDEBUG1( CS_INFO, "%s is UNPROTECTED.\n", szUrl );
			return( SF_STATUS_REQ_NEXT_NOTIFICATION );
		case PROTECTED :
			CSDEBUG1( CS_INFO, "%s is PROTECTED.\n", szUrl );
			break;
		case ALLOWUNAUTHENTICATEDACCESS :
			CSDEBUG1( CS_INFO, "%s is ALLOWUNAUTHENTICATEDACCESS.\n", szUrl );
			break;
		default :
			// Something bad happened!
			CosignLog( CS_ERROR, "Could not determine protected status of %s.  Got %d.\n", szUrl, protectedStatus );
			SetLastError( ERROR_INTERNAL_ERROR );
			return( SF_STATUS_REQ_ERROR );
	}
	CSDEBUG0( CS_INFO, "Setting COSIGN_SERVICE server variable.\n" );
	pfph->AddHeader( pfc, "COSIGN-SERVICE:", scd->scd_name );
	pfph->AddHeader( pfc, "COSIGN_SERVICE:", scd->scd_name );
	CSDEBUG0( CS_INFO, "Done setting COSIGN_SERVICE server variable.\n" );

	dwSize = DEFAULTBUFFERSIZE;
	if ( !pfph->GetHeader( pfc, "Cookie:", szBuffer, &dwSize ) ) {
		CosignLog( CS_INFO, "OnPreprocHeaders(): Could not get cookie.\n" );
		if ( dwSize > DEFAULTBUFFERSIZE ) {
			//Buffer was too small!
			CosignLog( CS_ERROR,  "OnPreprocHeaders(): buffer too small for cookie.\n" );
			SetLastError( ERROR_INTERNAL_ERROR );
			return( SF_STATUS_REQ_ERROR );
		}
		if ( protectedStatus == ALLOWUNAUTHENTICATEDACCESS) { 
			return( SF_STATUS_REQ_NEXT_NOTIFICATION );
		}	
		if ( SetCookieAndRedirect( pfc, pfph, scd ) != 0 ) {
			CosignLog( CS_ERROR, "OnPreprocHeaders(): An error occurred while trying to set the service cookie.\n" );
			SetLastError( ERROR_INTERNAL_ERROR );
            return( SF_STATUS_REQ_ERROR );
		}
		return( SF_STATUS_REQ_FINISHED );
	}

	//Look for a cookie for the protected service, if one isn't found, set and redirect
	CSDEBUG2( CS_INFO, "FindCookie()\nService Name = %s\nRaw cookie = %s\n", scd->scd_name, szBuffer );
	dwSize = DEFAULTCOOKIESIZE + 1;
	if ( FindCookie( scd->scd_name, szBuffer, szCookie, dwSize ) < 1 ) {
		if ( protectedStatus == ALLOWUNAUTHENTICATEDACCESS ) { 
			return( SF_STATUS_REQ_NEXT_NOTIFICATION );
		}	
		CSDEBUG0( CS_INFO, "OnPreprocHeaders(): Could not FindCookie(), redirecting\n" );
		SetCookieAndRedirect( pfc, pfph, scd );
		return( SF_STATUS_REQ_FINISHED );
	}
	//szCookie contains the cookie we want to validate
    CSDEBUG2( CS_INFO, "Got cookie for: %s\ndata=%s\n", scd->scd_name, szCookie );

    if ( !validchars( szCookie ) ) {
        CosignLog( CS_ERROR, "Invalid service cookie.  Setting new cookie.\n" );
        SetCookieAndRedirect( pfc, pfph, scd );
		return( SF_STATUS_REQ_FINISHED );
    }

	dwSize = DEFAULTCOOKIESIZE;
	if ( !pfc->GetServerVariable( pfc, "REMOTE_ADDR", szIpAddr, &dwSize ) ) {
		CosignLog( CS_ERROR, "Unable to get client's IP address.\n" );
		SetLastError( ERROR_INTERNAL_ERROR );
        return( SF_STATUS_REQ_ERROR );
	}
	
	EnterCriticalSection( &g_csConnPool );

	if ( (cp = GrabConnection( g_cp ) ) == NULL ) {
		///unable to grab connection!
		///This needs to be changed to dynamically increase the size of the connection pool.  For now, log the error.
		CosignLog( CS_ERROR, "Connection pool is busy, perhaps the current maximum of %n is too low?\n", g_cs.cs_connPoolSize );
		CSReportEvent( EVENTLOG_WARNING_TYPE, 0, CS_CONNECTIONPOOLBUSY, 0, NULL );
		LeaveCriticalSection( &g_csConnPool );
        SetLastError( ERROR_INTERNAL_ERROR );
		return( SF_STATUS_REQ_ERROR );
	}
	LeaveCriticalSection( &g_csConnPool );

	if ( g_ctx == NULL ) {
		CSDEBUG0( CS_INFO, "OnPreprocheaders, g_ctx is NULL.\n" );
	}
	cv = CookieValid( szCookie, &(cp->cp_cl), scd, &g_cs, &si, szIpAddr, g_ctx );

	EnterCriticalSection( &g_csConnPool );
	ReleaseConnection( cp );
	LeaveCriticalSection( &g_csConnPool );

	if ( cv == COSIGN_ERROR ) {
		CosignLog( CS_ERROR, "OnPreprocHeaders()::CookieValid() error.\n" );
		if ( protectedStatus == ALLOWUNAUTHENTICATEDACCESS) { 
			return( SF_STATUS_REQ_NEXT_NOTIFICATION );
		}
		SetLastError( ERROR_INTERNAL_ERROR );
		return( SF_STATUS_REQ_ERROR );
	}

    if ( cv == COSIGN_LOGGED_OUT ) {
        CSDEBUG0( CS_INFO, "User is logged out.  Setting cookie and redirecting!\n" );
        SetCookieAndRedirect( pfc, pfph, scd );
        return( SF_STATUS_REQ_FINISHED );
    }

    if ( cv == COSIGN_BAD_SUFFIX ) {
        SetLastError( ERROR_INTERNAL_ERROR );
        return( SF_STATUS_REQ_ERROR );
    }

	//Everything shines, let them through
	if ( cv == COSIGN_OK ) {

		if ( g_cs.cs_getkrbtkt ) {
			char *dir;
			CSDEBUG0( CS_INFO, "Doing some kerberos ticket stuff.\n" );
			if (g_cs.cs_krbTicketDir == NULL || *(g_cs.cs_krbTicketDir) == '\0') {
				dir = DEFAULT_TKT_PREFIX;
			} else {
				dir = g_cs.cs_krbTicketDir;
			} 
			if ( _snprintf( krbpath, sizeof(krbpath), "FILE:%s\\%s", dir, si.si_krb5tktfile ) > sizeof(krbpath)) {
                CosignLog( CS_ERROR, "OnPreprocHeaders(): _snprintf failed.\n" );
                SetLastError( ERROR_INTERNAL_ERROR );
                return( SF_STATUS_REQ_ERROR );
			}
			pfph->AddHeader( pfc, "KRB5CCNAME:", krbpath ); 
		}
		if ( g_cs.cs_getProxy && 
			 g_cs.cs_proxyDir != NULL &&
			 *(g_cs.cs_proxyDir) != '\0') {
			CSDEBUG0( CS_INFO, "Setting some proxy cookie stuff.\n" );
			if ( _snprintf( proxypath, sizeof( proxypath ), "%s\\%s=%s", g_cs.cs_proxyDir, scd->scd_name, szCookie ) > sizeof( proxypath )) {
				CosignLog( CS_ERROR, "OnPreprocHeaders(): proxy cookie path too long");
                SetLastError( ERROR_INTERNAL_ERROR );
                return( SF_STATUS_REQ_ERROR );
			}
			pfph->AddHeader( pfc, "PROXYFNAME:", proxypath );
		}

		//ASP likes dashes instead of underscores for some reason, hence the duplication of variables
		pfph->AddHeader( pfc, "REMOTE-USER:", si.si_user );
		pfph->AddHeader( pfc, "REMOTE-REALM:", si.si_realm );
        pfph->AddHeader( pfc, "COSIGN-FACTOR:", si.si_factor );
		
		//For all other 'normal' scripting languages
		pfph->AddHeader( pfc, "REMOTE_USER:", si.si_user );
		pfph->AddHeader( pfc, "REMOTE_REALM:", si.si_realm );
        pfph->AddHeader( pfc, "COSIGN_FACTOR:", si.si_factor );
		return ( SF_STATUS_REQ_NEXT_NOTIFICATION );
	} else if ( protectedStatus == ALLOWUNAUTHENTICATEDACCESS ) {
		return( SF_STATUS_REQ_NEXT_NOTIFICATION );
	}
	CSDEBUG0( CS_INFO, "OnPreprocHeaders(): Out of options.  Setting new cookie.\n" );
	SetCookieAndRedirect( pfc, pfph, scd );
	return( SF_STATUS_REQ_FINISHED );

}


/*
 * DllMain
 *
 * Standard Windows DLL entry point.
 *
 */
	BOOL WINAPI
DllMain( HINSTANCE hDll, DWORD dwReason, LPVOID lpReserved )
{

	switch( dwReason ) {
	case DLL_PROCESS_ATTACH :
		// DLL initialization here
		// Do not call Windows sockets functions here!
		return( TRUE );
	case DLL_PROCESS_DETACH :
		// DLL cleanup here
		break;
	case DLL_THREAD_ATTACH :
		// Thread specific initialization
		break;
	case DLL_THREAD_DETACH :
		// Thread specific cleanup
		break;
	default :
		// Reason unknown.  DLL caller must have a secret agenda.
		break;
	}

	return( TRUE );
}

/*
 * HttpFilterProc
 *
 * returns DWORD - Tells IIS how the request was handled.
 *
 * in variables - Tells filter the state of the browser/server connection.
 *
 * Standard, exported IIS filter entry function.
 *
 */
	DWORD WINAPI
HttpFilterProc( PHTTP_FILTER_CONTEXT pfc, DWORD dwNotificationType, LPVOID pvNotification )
{

	switch ( dwNotificationType ) {

	case SF_NOTIFY_PREPROC_HEADERS :
		return( OnPreprocHeaders( pfc, (PHTTP_FILTER_PREPROC_HEADERS)pvNotification ) );
	default :
		break;
	}
	return( SF_STATUS_REQ_NEXT_NOTIFICATION );
}


/*
 * GetFilterVersion
 *
 * returns BOOL - TRUE, filter loaded properly.  FALSE, there was an error
 * loading the filter.
 *
 * PHTTP_FILTER_VERSION pVer - Some useful information for IIS and Windows.
 *
 * Standard, exported IIS filter entry function.  This gets called first.  All
 * of the filter initialization is done from here: config file is loaded,
 * connection pool is created, SSL certificates are loaded, Winsock is started,
 * criticial sections intialized, notifications for the filter are requested, and
 * the log file is created.
 *
 */
	BOOL WINAPI
GetFilterVersion( PHTTP_FILTER_VERSION pVer )
{

	HKEY	hkConfig;
	BYTE	lpData[ DEFAULTBUFFERSIZE ];
	BYTE	winsockVersionHigh = 2;
	BYTE	winsockVersionLow = 2;
	DWORD	dwSize = 0;
	DWORD	dwType;
	DWORD	err;
	int		useEventViewer = 0;
	WSADATA	wsaData;

	CosignLog( CS_INFO, "GetFilterVersion: %s\n", CSINITMSG );
	g_cs.cs_runInDefaultMode = FALSE;
	pVer->dwFilterVersion = HTTP_FILTER_REVISION;
	pVer->dwFlags = ( SF_NOTIFY_SECURE_PORT |
		              SF_NOTIFY_NONSECURE_PORT |
					  SF_NOTIFY_PREPROC_HEADERS |
					  SF_NOTIFY_ORDER_HIGH );
	strcpy( pVer->lpszFilterDesc, "Cosign Filter" );

		try {

	//Find location and name of XML config file from the Registry
	if ( (err = RegOpenKeyEx( HKEY_LOCAL_MACHINE, CSCONFIGKEY, 0, KEY_READ, &hkConfig )) != ERROR_SUCCESS ) {
		CosignLog( err, "GetFilterVersion::RegOpenKeyEx( %s ) failed.\n", CSCONFIGKEY );
		throw( err );
	}

	dwSize = DEFAULTBUFFERSIZE;
	if ( (err = RegQueryValueEx( hkConfig, CSCONFIGFILE, 0, &dwType, lpData, &dwSize )) != ERROR_SUCCESS ) {
		CosignLog( err, "GetFilterVersion::RegQueryValueEx( %s )\n", CSCONFIGFILE );
		RegCloseKey( hkConfig );
		throw( err );
	}

	CSDEBUG0( CS_INFO, "Attempting to load config file.\n" );
	if ( LoadConfigFile( &g_sl, (char*)lpData, &g_cs ) != 0 ) {
		CosignLog( CS_INFO, "Failed to load XML configuration file.\n" );
		LPCTSTR	errAry[ 1 ] = { (char*)lpData };
		CSReportEvent( EVENTLOG_ERROR_TYPE, 0, CS_XMLLOADFAILED, 1, errAry );
		RegCloseKey( hkConfig );
		throw ( CS_XMLLOADFAILED );
	}
	RegCloseKey( hkConfig );
	CosignLog( CS_INFO, "Loaded config file.\n" );

	if ( strcmp( g_cs.cs_writeDataToEventViewer, "TRUE" ) == 0 ) {
		useEventViewer = 1;
	}
	//Set up logging method
	if ( InitLog( useEventViewer, g_cs.cs_logFileLevel, g_cs.cs_logFilePath, g_cs.cs_logFileSize, CSINITMSG ) < 0 ) {
		//Continuing without the log file is ok.
		LPCTSTR	errAry[ 1 ] = { (char*)g_cs.cs_logFilePath };
		CSReportEvent( EVENTLOG_WARNING_TYPE, 0, CS_INITLOGFAILED, 1, errAry );
	}

	//Start Windows Sockets
	if ( (err = (DWORD)WSAStartup( MAKEWORD( winsockVersionLow, winsockVersionHigh ), &wsaData ) ) != 0 ) {
		CosignLog( err, "WSAStartup() Failed.\n" );
		throw( err );
	}
	if ( LOBYTE( wsaData.wVersion ) != winsockVersionLow  ||
        HIBYTE( wsaData.wVersion ) != winsockVersionHigh  ) {
		CosignLog( CS_ERROR, "Didn't obtain the winsock version %d.%d needed.  Received version %d.%d\n",
			LOBYTE( wsaData.wVersion ), HIBYTE( wsaData.wVersion ), winsockVersionLow, winsockVersionHigh );
		WSACleanup();
		throw( WSAVERNOTSUPPORTED );
	}
	

	if ( CreateConnPool( &g_cs, &g_cp, g_cs.cs_server, g_cs.cs_connPoolSize ) < 1 ) {
        throw( (DWORD)WSAGetLastError() );
	}
	CSDEBUG0( CS_INFO, "Populated connection list.\n" );

	//Initialize SSL Certificates

#ifdef TLS
	CSDEBUG0( CS_INFO, "Getting certificates...\n" );
	if ( SetCosignCerts( &g_cs, &g_ctx ) != 0 ) {
		CosignLog( CS_ERROR, "GetFilterVersion::SetCosignCerts() failed\n" );
        CSDEBUG0( CS_INFO, "Setting default mode to true.\n" );
			g_cs.cs_runInDefaultMode = TRUE;
            CSDEBUG0( CS_INFO, "CSReportEvent( EVENTLOG_ERROR_TYPE, 0, CS_DEFAULTMODE, 0, NULL );\n" );
			CSReportEvent( EVENTLOG_ERROR_TYPE, 0, CS_DEFAULTMODE, 0, NULL );
			CSDEBUG0( CS_INFO, "GetFilterVersion returning TRUE after setting run indefault mode to TRUE.\n" );
            return( TRUE );
	}
	if ( g_ctx == NULL ) {
		CSDEBUG0( CS_INFO, "GetFilterVersion, g_ctx is NULL.\n" );
	}
	CSDEBUG0( CS_INFO, "Done getting SSL certificates.\n" );
#endif


	if ( InitializeCriticalSectionAndSpinCount( &g_csConn, 0 ) == 0 ) {
		err = GetLastError();
		CosignLog( err, "Could not initialize the single connection critical section.\n" );
		throw( err );
	}
	if ( InitializeCriticalSectionAndSpinCount( &g_csConnPool, 0 ) == 0 ) {
		err = GetLastError();
		CosignLog( err, "Could not initialize Connection Pool critical section.\n" );
		throw( err );
	}
		} catch ( DWORD dwErr ) {

            CSDEBUG0( CS_INFO, "Setting default mode to true.\n" );
			g_cs.cs_runInDefaultMode = TRUE;
            CSDEBUG0( CS_INFO, "CSReportEvent( EVENTLOG_ERROR_TYPE, 0, CS_DEFAULTMODE, 0, NULL );\n" );
			CSReportEvent( EVENTLOG_ERROR_TYPE, 0, CS_DEFAULTMODE, 0, NULL );
			CSDEBUG0( CS_INFO, "GetFilterVersion returning TRUE after setting run indefault mode to TRUE.\n" );
            return( TRUE );

		}
	LPCTSTR		strTemp[ 1 ] = { VERSION };
	CSReportEvent( EVENTLOG_INFORMATION_TYPE, 0, CS_STARTED, 1, strTemp );	

	return( TRUE );
}

/*
 * TerminateFilter
 *
 * returns BOOL - TRUE, filter cleanup was successful.  FALSE, there was an error
 * cleaning something.
 *
 * DWORD dwFlags - From MS documenation: "No values for dwFlags have been identified at this time."
 * If you identify one, please contact cosign@umich.edu and probably Microsoft, too.  I'm sure
 * they'd want to know about it.
 *
 * The third, necessary, exported function for a proper filter DLL.
 *
 */
BOOL WINAPI
TerminateFilter( DWORD dwFlags )
{

	CosignLog( CS_INFO, "Unloading Cosign Filter.\n" );
    
	//Close sockets.  Shutdown Winsock.  Free SSL Certificates?

	UnloadConfigFile( &g_cs );
	ReleaseConnPool( g_cp );
	DeleteCriticalSection( &g_csConn );
	DeleteCriticalSection( &g_csConnPool );
	CSDEBUG0( CS_INFO, "Closing log file.\n" );
	CloseLog();
	CSDEBUG0( CS_INFO, "Done closing log file.\n" );

	WSACleanup();

	return( TRUE );
}

