/*
 * Copyright (c) 2003 - 2005 Regents of The University of Michigan.
 * All Rights Reserved.  See COPYRIGHT.
 */

#define INITGUID 
#define DEFAULT_BUFFER_SIZE 1024
#define _WIN32_DCOM 
#include <Windows.h>

#include <initguid.h>
#include <CGUID.H>
#include <iadmw.h>  // COM Interface header file. 
#include <iiscnfg.h>  // MD_ & IIS_MD_ #defines header file.
#include <atlBase.h>  // ATL support header file.

#include <httpfilt.h>
#include <objbase.h>
#include <msxml2.h>
#include <comutil.h>

#include <openssl\ssl.h>
#include <openssl\err.h>
#include <libsnet\snet.h>

#include "log.h"
#include "cosignmsg.h"
#include "cosignstructs.h"
#include "config.h"

typedef vector<FOLDERDATA> STRVECTOR;
#define COOKIESECURE ";secure"
#define COOKIEHTTPONLY ";httponly"

int
getNestedElements( IXMLDOMNodeList*	pXNL, long  items, BSTR ary[] );
int
newGetNestedElement( IXMLDOMNodeList*	pXNL, COSIGNSETTINGS* pcs  );

/* We're only concerned with the first instance.  All other
 * elements with this name will be ignored.
 */
HRESULT
GetConfigItemData( IXMLDOMDocument2* pDoc, WCHAR* item, VARIANT* data ) {

    IXMLDOMNodeList*	pXNL;
    IXMLDOMNode*		pXDN;
    IXMLDOMNode*        pXDNText;
    HRESULT             hr;

    if ( (hr = pDoc->getElementsByTagName( item, &pXNL )) != S_OK ) return( hr );
    if ( (hr = pXNL->get_item( 0, &pXDN )) != S_OK ) return( hr );
    if ( (hr = pXDN->get_firstChild( &pXDNText )) != S_OK ) return( hr );
    if ( (hr = pXDNText->get_nodeValue( data )) != S_OK ) return( hr );

    pXDN->Release();
    pXNL->Release();
    pXDNText->Release();

    return( hr );
}

void
UnloadConfigFile( COSIGNSETTINGS* pcs ) {

    free( pcs->cs_server );
	free( pcs->cs_cookieDBPath );
	free( pcs->cs_logFilePath );
	free( pcs->cs_writeDataToEventViewer );
	free( pcs->cs_CAFilePath );
	free( pcs->cs_chainFilePath );
    free( pcs->cs_privateKeyFilePath );
	free( pcs->cs_redirectURL );
    free( pcs->cs_postErrorRedirectURL );
	free( pcs->cs_krbTicketDir );
}

void
LoadDefaultSettings( COSIGNSETTINGS* pcs ) {
	
    pcs->cs_runInDefaultMode = FALSE;
    pcs->cs_server = NULL;
    pcs->cs_cookieDBPath = NULL;
    pcs->cs_logFilePath = NULL;
    pcs->cs_writeDataToEventViewer = NULL;
    pcs->cs_connPoolSize = 0;
    pcs->cs_cookieDBExpireTime = -1;
    pcs->cs_logFileSize = DEFAULTLOGFILESIZE;
#ifdef _DEBUG
    pcs->cs_logFileLevel = LOGGINGDEBUG;
#else
    pcs->cs_logFileLevel = LOGGINGNORMAL;
#endif
    pcs->cs_checkIpAddr = never;
    pcs->cs_redirectToHttps = FALSE;
	pcs->cs_cookieHttpOnly = FALSE;
	pcs->cs_cookieSecure = TRUE;
	pcs->cs_cookieOptions = NULL;
	pcs->cs_CAFilePath = NULL;
	pcs->cs_chainFilePath = NULL;
	pcs->cs_privateKeyFilePath = NULL;
    pcs->cs_redirectURL = NULL;
    pcs->cs_postErrorRedirectURL = NULL;
    pcs->cs_port = DEFAULTPORT;
    pcs->cs_getkrbtkt = FALSE;
	pcs->cs_krbTicketDir = NULL;
    pcs->cs_protocolVersionNeeded = 0;
    pcs->cs_getProxy = FALSE;
 	pcs->cs_proxyDir = NULL;
	pcs->cs_hashlen = 128;
}

DWORD
GetIISServerInstance( WCHAR*	webSiteName ) {

	HRESULT	hr, hResult;
	DWORD	index = 0;
	DWORD	dataIndex = 0;

	METADATA_HANDLE mhMetabase; 
	WCHAR	path[ METADATA_MAX_NAME_LEN ];
	WCHAR	subPath[ METADATA_MAX_NAME_LEN ];
	WCHAR	mdName[ METADATA_MAX_NAME_LEN ];
	CComPtr <IMSAdminBase> pIMeta; 

	METADATA_RECORD	mdr;
	DWORD dwBufSize = DEFAULT_BUFFER_SIZE; 
	DWORD dwReqBufSize = 0; 
	PBYTE pbBuf = new BYTE[dwBufSize]; 

	hr = CoInitialize( NULL );
	hr = CoCreateInstance(CLSID_MSAdminBase, NULL, CLSCTX_ALL, 
		IID_IMSAdminBase, (void **) &pIMeta); 

	swprintf( path, METADATA_MAX_NAME_LEN, L"/LM/W3SVC/" );
	for( index = 0;
		 SUCCEEDED( (hResult = pIMeta->EnumKeys( METADATA_MASTER_ROOT_HANDLE, path, (LPWSTR)&mdName, index )) );
		 index++ ) {
			///This swprintf() does NOT work.  Have no idea why.
			///Unfortunately, this means a use of a string copy and cat.  Ah well.
			///swprintf( subPath, L"%S%S", path, mdName );
			wcsncpy( subPath, path, METADATA_MAX_NAME_LEN );
			wcsncat( subPath,mdName, METADATA_MAX_NAME_LEN );
			CSDEBUG3( CS_INFO, "name=%S\npath=%S\nsubPath=%S\n", mdName, path, subPath );
			if (!SUCCEEDED( pIMeta->OpenKey( METADATA_MASTER_ROOT_HANDLE, subPath, METADATA_PERMISSION_READ, 20, &mhMetabase )) ) {
				//If we're unable to open a key at this level, assume it wasn't a web site
				continue;
				CSDEBUG0( CS_INFO, "pIMEta->OpenKey() failed.\n" );
			}
			mdr.dwMDAttributes = METADATA_INHERIT;
			mdr.dwMDUserType = IIS_MD_UT_SERVER;
			mdr.dwMDDataType = ALL_METADATA;
			mdr.dwMDDataLen = dwBufSize;
			mdr.pbMDData = pbBuf;
			mdr.dwMDIdentifier = MD_SERVER_COMMENT;

			if (SUCCEEDED(pIMeta->GetData( mhMetabase, L"/", &mdr, &dwReqBufSize )) ) {
				if ( wcscmp( (WCHAR*)mdr.pbMDData, webSiteName ) == 0 ) { 
					//This is the success condition.
					pIMeta->CloseKey( mhMetabase );
					return( _wtol( mdName ) );
				}
			}
			//if no data with this label, assume we don't care about it and carry on.
			pIMeta->CloseKey( mhMetabase );
	}
	//A web site description, "ServerComment", could not be found.
	return( -1 );
}

/*
 * LoadConfigFile
 *
 * returns - 0 if everything went ok
 *         - nonzero means something bad happened.
 *         - This will probably be changed to corresponed to message defined in
 *           CosignMsg.h
 *
 * SERVICELIST* - Service list point that is populated by this function.
 * char*		- Location of configuration file.
 * COSIGNSETTINGS* - Variable that keeps track of the cosign settings.
 *                   Populated based on the contents of the configuration file.
 */
int
LoadConfigFile( SERVICELIST *sl, char* szFileName, COSIGNSETTINGS* pcs )
{

	HRESULT hr;
	IXMLDOMDocument2*	pXMLDoc = NULL;
	IXMLDOMParseError*	pXMLParseErr = NULL; 
	IXMLDOMNodeList*	pXNL = NULL;
    IXMLDOMNodeList*	pXNLChildren = NULL;
	IXMLDOMNode*		pXDN = NULL;
	IXMLDOMNode*		pXDNElem = NULL;
    IXMLDOMNode*		pXDNAttribute = NULL;
    IXMLDOMNode*        pXDNText = NULL;
    IXMLDOMNode*        pXDNChild = NULL;
	IXMLDOMNamedNodeMap* pXNM = NULL;
    IXMLDOMParseError*  pXDPE = NULL;
    DOMNodeType         dNT;
    long                error, line, numServices, cur, numChildren;
	BSTR bstr, bstrBaseName;
	VARIANT varDoc;
	VARIANT_BOOL varStatus;
    VARIANT v;
	int	ret = 0;
    int protocolVersionNeeded = 0;
	int	items;
	int	cookieOptionsSize;
    long curChild;


	SERVICECONFIGDATA*	scd;

	CSDEBUG0( CS_INFO, "Attempting to load default settings.\n" );
	LoadDefaultSettings( pcs );
	CSDEBUG0( CS_INFO, "Default settings loaded.\n" );

	hr = CoInitializeEx( NULL, COINIT_MULTITHREADED );
	if ( hr == S_OK || hr == S_FALSE ) {
		//everything's cool
	} else if ( hr == RPC_E_CHANGED_MODE ) {
		CosignLog( hr, "CoInitializeEx() could not change thread mode.  Continuing anyway.\n" );
	} else {
		CosignLog( hr, "CoInitializeEx() failed.\n" );
		return( -1 );
	}
	if ( (hr = CoCreateInstance(CLSID_DOMDocument40, NULL, CLSCTX_INPROC_SERVER, 
		IID_IXMLDOMDocument2, (void**)&pXMLDoc) ) != S_OK ) {
		CosignLog( hr, "CoCreateInstance() failed.\n" );
		CoUninitialize();
		return( -1 );
	}
	VariantInit( &varDoc );
	V_BSTR(&varDoc) = _com_util::ConvertStringToBSTR( szFileName );
	V_VT(&varDoc) = VT_BSTR;

	try {

	if ( (hr = pXMLDoc->put_async( VARIANT_FALSE ) ) != S_OK ) {
		throw "pXMLDoc->put_async() failed\n";
	}
	if ( (hr = pXMLDoc->put_validateOnParse( VARIANT_TRUE ) ) != S_OK ) {
		throw "pXMLDoc->put_validateOnParse() failed.\n";
	}
	if ( (hr = pXMLDoc->put_resolveExternals( VARIANT_TRUE ) ) != S_OK ) {
		throw "pXMLDoc->put_resolveExternals() failed.\n";
	}
	if ( (hr = pXMLDoc->load( varDoc, &varStatus ) ) != S_OK ) {
		throw "pXMLDoc->load() failed.\n";
	}
	if ( (hr = pXMLDoc->get_xml(&bstr) ) != S_OK ) {
		throw "pXMLDoc->get_xml() failed.\n";
	}

    if ( (hr = pXMLDoc->validate( &pXDPE )) != S_OK ) {
        pXDPE->get_errorCode( &error );
        pXDPE->get_reason( &bstr );
        pXDPE->get_line( &line );
        CSDEBUG3( CS_INFO, "Parser error code at line %u, %u\n%S", line, error, bstr );
        throw "pXMLDoc->validate() failed.\n";
    }
	CSDEBUG0( CS_INFO, "Loaded XML file succesfully.  Parsing...\n" );
		
	} catch( char* szDesc ) {
        CSDEBUG0( CS_INFO, szDesc );
        pXMLDoc->get_parseError( &pXMLParseErr );
		if ( pXMLParseErr ) {
            BSTR	bstrReason;
            char    lineNumber[ 512 ];
			pXMLParseErr->get_reason( &bstrReason );
            pXMLParseErr->get_line( &line );
			CSDEBUG2( CS_INFO, "Parsing error at line %u: %S.\n", line, bstrReason );
            _snprintf( lineNumber, 512, "%u", line );
            LPCTSTR	errAry[ 2 ] = { lineNumber, (char*)_com_util::ConvertBSTRToString( bstrReason ) };
		    CSReportEvent( EVENTLOG_ERROR_TYPE, 0, CS_XMLPARSEERROR, 2, errAry );
		}		
		pXMLParseErr->Release();
		ret--;
        return( ret );
	}  
    VariantInit( &v );
    pcs->cs_runInDefaultMode = TRUE;
	
    GetConfigItemData( pXMLDoc, L"CAFilePath", &v );
    pcs->cs_CAFilePath = _com_util::ConvertBSTRToString( V_BSTR(&v) );
    CSDEBUG1( CS_INFO, "CAFilePath = %s\n", pcs->cs_CAFilePath );

    GetConfigItemData( pXMLDoc, L"ChainFilePath", &v );
    pcs->cs_chainFilePath = _com_util::ConvertBSTRToString( V_BSTR(&v) );
    CSDEBUG1( CS_INFO, "ChainFilePath = %s\n", pcs->cs_chainFilePath );

    GetConfigItemData( pXMLDoc, L"PrivateKeyFilePath", &v );
    pcs->cs_privateKeyFilePath = _com_util::ConvertBSTRToString( V_BSTR(&v) );
    CSDEBUG1( CS_INFO, "PrivateKeyFilePath = %s\n", pcs->cs_privateKeyFilePath );

    // <Log>
    //	<Path>C:\Program Files\Cosign\Logs</Path>
	//	<MaxSizeKB>4096</MaxSizeKB>
	//	<WriteDataToEventViewer>false</WriteDataToEventViewer>
	// </Log>
    BSTR    ary[ 4 ];
    pXMLDoc->getElementsByTagName( L"Log", &pXNL );
    getNestedElements( pXNL, 3, ary );
    pcs->cs_logFilePath = _com_util::ConvertBSTRToString( ary[ 0 ] );
    pcs->cs_logFileSize = atoi( _com_util::ConvertBSTRToString( ary[ 1 ] ) ) * 1024;
    pcs->cs_writeDataToEventViewer = _com_util::ConvertBSTRToString( ary[ 2 ] );
    pXNL->Release();


    pXMLDoc->getElementsByTagName( L"CookieDB", &pXNL );
    getNestedElements( pXNL, 2, ary );
    pcs->cs_cookieDBPath = _com_util::ConvertBSTRToString( ary[ 0 ] );
    pcs->cs_cookieDBExpireTime = atoi( (_com_util::ConvertBSTRToString( ary[ 1 ] )) );
    pcs->cs_cookieDBExpireTime *= 10000000;
    CSDEBUG1( CS_INFO, "pcs->cs_cookieDBExpireTime = %d\n", pcs->cs_cookieDBExpireTime );
    pXNL->Release();


	pcs->cs_cookieSecure = TRUE;
	pcs->cs_cookieHttpOnly = FALSE;
	pcs->cs_cookieOptions = NULL;
	// "\r\n" + '\0'
	// See SetCookieAndRedirect()
	cookieOptionsSize = 3 + strlen(COOKIESECURE) + strlen(COOKIEHTTPONLY);
	pcs->cs_cookieOptions = (char*)malloc( cookieOptionsSize );
	items = 0;
	hr = pXMLDoc->getElementsByTagName( L"Cookies", &pXNL );
	if ( SUCCEEDED( hr ) ) {
		items = newGetNestedElement( pXNL, pcs );
	}
	if ( pcs->cs_cookieSecure || pcs->cs_cookieHttpOnly ) {
		CSDEBUG1( CS_INFO, "sizeof( pcs->cs_cookieOptions) = %d\n", sizeof( *pcs->cs_cookieOptions) );
		if ( pcs->cs_cookieSecure) {
			CSDEBUG0( CS_INFO,"_snprintf()'ing for COOKIESECURE\n" );
			_snprintf( pcs->cs_cookieOptions, cookieOptionsSize, "%s", COOKIESECURE );
		}
		if ( pcs->cs_cookieHttpOnly ) {
			CSDEBUG0( CS_INFO,"_snprintf()'ing or COOKIEHTTPONLY\n" );
			_snprintf( pcs->cs_cookieOptions, cookieOptionsSize, "%s%s", pcs->cs_cookieOptions, COOKIEHTTPONLY);
		}
		CSDEBUG0( CS_INFO, "_snprintf()'ing \\r\\n\n" );
		_snprintf( pcs->cs_cookieOptions, cookieOptionsSize, "%s\r\n", pcs->cs_cookieOptions );
	} else {
		_snprintf( pcs->cs_cookieOptions, cookieOptionsSize, "\r\n" );
	}
	CSDEBUG1( CS_INFO, "Cookies will be set to %s\n", pcs->cs_cookieOptions );


    pXMLDoc->getElementsByTagName( L"LoginServer", &pXNL );
    getNestedElements( pXNL, 4, ary );
    pcs->cs_server = _com_util::ConvertBSTRToString( ary[ 0 ] );
    pcs->cs_redirectURL = _com_util::ConvertBSTRToString( ary[ 1 ] );
    pcs->cs_postErrorRedirectURL = _com_util::ConvertBSTRToString( ary[ 2 ] );
    pcs->cs_port = atoi( _com_util::ConvertBSTRToString( ary[ 3 ] ) );
    pXNL->Release();
    CSDEBUG1( CS_INFO, "LoginServer = %s\n", pcs->cs_server );

    if ( GetConfigItemData( pXMLDoc, L"ConnectionPoolSize", &v ) == S_OK ) {
        pcs->cs_connPoolSize = atoi( _com_util::ConvertBSTRToString( V_BSTR(&v) ) );
    }
    CSDEBUG1( CS_INFO, "pcs->connPoolSize = %d\n", pcs->cs_connPoolSize );

    pcs->cs_redirectToHttps = TRUE;
    if ( (GetConfigItemData( pXMLDoc, L"HttpsOnly", &v ) == S_OK ) &&
        (_wcsicmp( V_BSTR(&v), L"false" ) == 0 ) ) {
        pcs->cs_redirectToHttps = FALSE;
    }
    CSDEBUG1( CS_INFO,  "pcs->httpsOnly= %d\n", pcs->cs_redirectToHttps );

    pcs->cs_checkIpAddr = never;
    if ( (GetConfigItemData( pXMLDoc, L"CheckIPAddress", &v ) == S_OK) ) {
        if ( (_wcsicmp( V_BSTR(&v), L"never" ) == 0 ) ) {
		    pcs->cs_checkIpAddr = never;
        } else if ( (_wcsicmp( V_BSTR(&v), L"initial" ) == 0 ) ) {
            pcs->cs_checkIpAddr = initial;
        } else if ( (_wcsicmp( V_BSTR(&v), L"always" ) == 0 ) ) {
            pcs->cs_checkIpAddr = always;
        } else {
            CosignLog( CS_ERROR, "LoadConfigFile: CheckIPAddress in an unknown state.  Maybe iiscosign.xsd is the wrong version?\n");
            ///should exit here
        }
    }
    CSDEBUG1( CS_INFO,  "CheckIPAddr = %d\n", pcs->cs_checkIpAddr );
    
    if ( GetConfigItemData( pXMLDoc, L"KerberosTicketsPath", &v ) != S_OK ) {
        pcs->cs_getkrbtkt = FALSE;
        pcs->cs_krbTicketDir = NULL;
    } else {
        pcs->cs_getkrbtkt = TRUE;
        pcs->cs_krbTicketDir = _com_util::ConvertBSTRToString( V_BSTR(&v) );
    }
    CSDEBUG1( CS_INFO,  "KerberosTicketsPath= %s\n", pcs->cs_krbTicketDir );

    if ( GetConfigItemData( pXMLDoc, L"ProxyCookiesPath", &v ) != S_OK ) {
        pcs->cs_getProxy = FALSE;
        pcs->cs_proxyDir = NULL;
    } else {
        pcs->cs_getProxy = TRUE;
        pcs->cs_proxyDir = _com_util::ConvertBSTRToString( V_BSTR(&v) );
    }
    CSDEBUG1( CS_INFO,  "ProxyCookiesPath= %s\n", pcs->cs_proxyDir );
     if ( GetConfigItemData( pXMLDoc, L"HashLength", &v ) == S_OK ) {
        pcs->cs_hashlen = atoi( _com_util::ConvertBSTRToString( V_BSTR(&v) ) );
    }
    CSDEBUG1( CS_INFO, "pcs->cs_hashlen = %d\n", pcs->cs_hashlen );


    pXMLDoc->getElementsByTagName( L"Service", &pXNL );
    pXNL->get_length( &numServices );
    for ( cur = 0; cur < numServices; cur++ ) {
        scd = new SERVICECONFIGDATA;
        pXNL->get_item( cur, &pXDN );
        pXDN->get_childNodes( &pXNLChildren );

        pXNLChildren->get_length( &numChildren );
        for ( curChild = 0; curChild < numChildren; curChild++ ) {
            pXNLChildren->get_item( curChild, &pXDNChild );
            if ( pXDNChild->get_nodeType( &dNT ) != S_OK ) {
                ///continue?
            }
            if ( dNT != NODE_ELEMENT ) {
                pXDNChild->Release();
                continue;
            }
            
            pXDNChild->get_text( &bstr );
            pXDNChild->get_baseName( &bstrBaseName );
            if ( _wcsicmp( L"Protected", bstrBaseName ) == 0 ) {
                scd->addURI( &bstr, PROTECTED );
                CSDEBUG1( CS_INFO, "%S is protected.\n", bstr );
            } else if ( _wcsicmp( L"AllowPublicAccess", bstrBaseName ) == 0 ) {
                scd->addURI( &bstr, ALLOWUNAUTHENTICATEDACCESS );
                CSDEBUG1( CS_INFO, "%S is allowpublicaccess.\n", bstr );
            } else if ( _wcsicmp( L"Unprotected", bstrBaseName ) == 0 ) {
                scd->addURI( &bstr, UNPROTECTED );
                CSDEBUG1( CS_INFO, "%S is unprotected.\n", bstr );
            } else if ( _wcsicmp( L"RequireFactor", bstrBaseName ) == 0 ) {
                protocolVersionNeeded++;
                scd->addFactor( &bstr );
            } else if ( _wcsicmp( L"FactorSuffix", bstrBaseName ) == 0 ) {
                protocolVersionNeeded++;
                scd->addIgnoreSuffix( &bstr );
            } else if ( _wcsicmp( L"IISDescription", bstrBaseName ) == 0 ) {
                scd->scd_ServerInstance = GetIISServerInstance( bstr );
                if ( scd->scd_ServerInstance == -1 ) {
				    LPCTSTR	errAry[ 1 ] = { (char*)_com_util::ConvertBSTRToString( bstr ) };
				    CSReportEvent( EVENTLOG_ERROR_TYPE, 0, CS_WEBSITENOTFOUND, 1, errAry );
                    CosignLog( CS_ERROR,  "Unable to find the web site \"%S\" on this server.  Check your XML configuration file.\n", bstr );
			    }
            } else if ( _wcsicmp( L"Website", bstrBaseName ) == 0 ) {
                scd->scd_webServer = _com_util::ConvertBSTRToString( bstr );
            } else if ( _wcsicmp( L"Name", bstrBaseName ) == 0 ) {
                scd->scd_name = _com_util::ConvertBSTRToString( bstr );
            } else if ( _wcsicmp( L"FactorIgnoreSuffixOn", bstrBaseName ) == 0 ) {
                if ( _wcsicmp( L"true", bstr ) == 0 ) {
                    CSDEBUG0( CS_INFO, "Setting scd_ignoreSuffixOn to TRUE\n" );
                    scd->scd_ignoreSuffixOn = TRUE;
                } else {
                    CSDEBUG0( CS_INFO, "Setting scd_ignoreSuffixOn to FALSE\n" );
                    scd->scd_ignoreSuffixOn = FALSE;
                }
            } else if ( _wcsicmp( L"SiteEntry", bstrBaseName ) == 0 ) {
                scd->scd_siteEntry = _com_util::ConvertBSTRToString( bstr );
            } else {
                // This could be caused by a bad iiscosign.xsd schema file.
                CosignLog( CS_ERROR, "Unrecognized config options \"%S\" is set to \"%S\".\n", bstrBaseName, bstr );
            }
            CSDEBUG2( CS_INFO,  "%S = %S\n", bstrBaseName, bstr );

            pXDNChild->Release();
        }
        pXNLChildren->Release();
        pXDN->Release();
        sl->addService( scd );
    }

    
	if ( &varDoc ) VariantClear( &varDoc );
	if ( pXMLDoc ) pXMLDoc->Release();
	CoUninitialize();

    // Multi-factor requires version 2.
    // If version 2 is not needed and the login server does not speak version 2, then the default
    // version 0 will be used.
    if ( protocolVersionNeeded ) {
        pcs->cs_protocolVersionNeeded = 2;
    } else {
        pcs->cs_protocolVersionNeeded = 0;
    }
    pcs->cs_runInDefaultMode = FALSE;

	return( ret );
}

int
getNestedElements( IXMLDOMNodeList*	pXNL, long  items, BSTR ary[] ) {

    IXMLDOMNodeList*	pXNLChildren = NULL;
	IXMLDOMNode*		pXDN = NULL;
    IXMLDOMNode*        pXDNChild = NULL;
    DOMNodeType         dNT;
	BSTR				bstrName;
    long                curItem, cur, numChildren;
	HRESULT				hr;

	if ( pXNL == NULL ) {
		return( 0 );
	}
    hr = pXNL->get_item( 0, &pXDN );
	if ( FAILED( hr ) || pXDN == NULL ) {
		return( 0 );
	}
    hr = pXDN->get_childNodes( &pXNLChildren );
	if ( FAILED( hr ) || pXNLChildren == NULL ) {
		return( 0 );
	}
    curItem = 0;
    cur = 0;
    for ( cur = 0, pXNLChildren->get_length( &numChildren );cur < numChildren && curItem < items; cur++ ) {
        pXNLChildren->get_item( cur, &pXDNChild );
        if ( pXDNChild->get_nodeType( &dNT ) != S_OK ) {
            ///continue?
        }
        
        if ( dNT != NODE_ELEMENT ) {
            pXDNChild->Release();
            continue;
        }
		pXDNChild->get_baseName( &bstrName );
        pXDNChild->get_text( &(ary[ curItem ]) );
        CSDEBUG2( CS_INFO, "Got element %S=\"%S\"!\n", bstrName, ary[ curItem ] );
        curItem++;
        pXDNChild->Release();
    }
    pXNLChildren->Release();
	return( numChildren );
}


int
newGetNestedElement( IXMLDOMNodeList*	pXNL, COSIGNSETTINGS* pcs ) {
	IXMLDOMNodeList*	pXNLChildren = NULL;
	IXMLDOMNode*		pXDN = NULL;
    IXMLDOMNode*        pXDNChild = NULL;
    DOMNodeType         dNT;
	BSTR				bstrName, bstrValue;
    long                curItem, cur, numChildren;
	HRESULT				hr;

	if ( pXNL == NULL ) {
		return( 0 );
	}
    hr = pXNL->get_item( 0, &pXDN );
	if ( FAILED( hr ) || pXDN == NULL ) {
		return( 0 );
	}
    hr = pXDN->get_childNodes( &pXNLChildren );
	if ( FAILED( hr ) || pXNLChildren == NULL ) {
		return( 0 );
	}
    curItem = 0;
    cur = 0;
    for ( cur = 0, pXNLChildren->get_length( &numChildren );cur < numChildren; cur++ ) {
        pXNLChildren->get_item( cur, &pXDNChild );
        if ( pXDNChild->get_nodeType( &dNT ) != S_OK ) {
            ///continue?
        }
        
        if ( dNT != NODE_ELEMENT ) {
            pXDNChild->Release();
            continue;
        }
		pXDNChild->get_baseName( &bstrName );
		pXDNChild->get_text( &bstrValue );
		if ( _wcsicmp( L"Secure", bstrName ) == 0 ){
			if ( _wcsicmp( L"true", bstrValue ) == 0 ) {
				pcs->cs_cookieSecure = TRUE;
			} else {
				pcs->cs_cookieSecure = FALSE;
			}
		} else if ( _wcsicmp( L"HttpOnly", bstrName ) == 0 ) {
			if ( _wcsicmp( L"true", bstrValue ) == 0 ) {
				pcs->cs_cookieHttpOnly = TRUE;
			} else {
				pcs->cs_cookieHttpOnly = FALSE;
			}
		} else {
			pXDNChild->Release();
			pXNLChildren->Release();
			CosignLog( CS_ERROR, "Found unexpected element in config file: %S=%S\n", bstrName, bstrValue );
			return( -1 );
		}
        
        CSDEBUG2( CS_INFO, "Got element %S=\"%S\"!\n", bstrName, bstrValue );
        curItem++;
        pXDNChild->Release();
    }
    pXNLChildren->Release();
	return( numChildren );
}
