
#include <vector>
#include <algorithm>
#include <comutil.h>
#include <string.h>
#include "pathcmp.h"

using namespace std;

#define UNPROTECTED -1
#define PROTECTED	1
#define ALLOWUNAUTHENTICATEDACCESS 0

class FACTOR {

public:
    FACTOR() {
        factor = NULL;
    }
    FACTOR( char* str ) {
        if ( str != NULL ) factor = (char *)malloc( strlen( str ) + 1 );
        if ( factor != NULL ) strcpy( factor, str );
    }
    ~FACTOR() {
        if ( factor ) free ( factor );
    }

    FACTOR( const FACTOR &f ) {
        *this = f;
    }

    FACTOR & operator=( const FACTOR &f ) {
        if ( this != &f ) {
            this->factor = (char*)malloc( strlen( f.factor ) + 1 );
            strcpy( this->factor, f.factor );
        }
        return( *this );
    }

    bool operator==( const FACTOR &f ) {
        if ( _stricmp( this->factor, f.factor ) == 0 ) {
            return true;
        }
        return false;
    }

private:
    char*   factor;
};

class IGNORESUFFIX {

public :
    IGNORESUFFIX() {
        suffix = NULL;
    }
    IGNORESUFFIX( char* str ) {
        if ( str != NULL ) suffix = (char *)malloc( strlen( str ) + 1 );
        if ( suffix != NULL ) strcpy( suffix, str );
    }
    ~IGNORESUFFIX() {
        if ( suffix ) free ( suffix );
    }

    IGNORESUFFIX( const IGNORESUFFIX &is ) {
        *this = is;
    }

    IGNORESUFFIX & operator=( const IGNORESUFFIX &is ) {
        if ( this != &is ) {
            this->suffix = (char*)malloc( strlen( is.suffix ) + 1 );
            strcpy( this->suffix, is.suffix );
        }
        return( *this );
    }

    bool operator==( const IGNORESUFFIX &is ) {
        
        if ( strlen( this->suffix ) > strlen( is.suffix ) ) {
            return false;
        }
        char * cmp;
        cmp = &(is.suffix[ strlen( is.suffix ) - strlen (this->suffix) ]);
        
        if ( _stricmp( this->suffix, cmp ) == 0 ) return true;
        return false;
    }
private:
    char*   suffix;
    friend class  SERVICECONFIGDATA;

};

class FOLDERDATA  {

public:

	char*	name;
	BOOL	protectSubfolders;
	int     protectedStatus;

	FOLDERDATA() {
		name = NULL;
        protectedStatus = PROTECTED;
		protectSubfolders = TRUE;
	}
	FOLDERDATA( char* str ) {
		//NOTE: Making a copy to be nice to caller.
		if ( str != NULL ) name = (char *)malloc( strlen( str ) + 1 );
		if ( name != NULL ) strcpy( name, str );
		protectedStatus = PROTECTED;
		protectSubfolders = TRUE;
	}
	FOLDERDATA( BSTR*	bstr ) {
		name = _com_util::ConvertBSTRToString( *bstr );
		protectedStatus = PROTECTED;
		protectSubfolders = TRUE;
	}
	FOLDERDATA( BSTR*	bstr, int ps ) {
		name = _com_util::ConvertBSTRToString( *bstr );
		protectedStatus = ps;
		protectSubfolders = TRUE;
	}
	~FOLDERDATA() {
		if ( name ) free (name );
	}
	//Copy Constructor
	FOLDERDATA( const FOLDERDATA &fd ) {
		*this = fd;
	}
	FOLDERDATA & operator=( const FOLDERDATA &fd ) {

		if ( this != &fd ) {
			this->name = (char*)malloc( (strlen( fd.name ) + 1) );
			strcpy( this->name, fd.name );
			this->protectSubfolders = fd.protectSubfolders;
            this->protectedStatus = fd.protectedStatus;
		}
		return( *this );
	}
	bool operator==( const FOLDERDATA &fd ) {

		if ( _stricmp( this->name, fd.name ) == 0 || ( _stricmp( this->name, "/" ) == 0 ) ) return true;
		if ( (strlen( fd.name ) > strlen( this->name ) ) &&
			 (fd.name[ strlen( this->name ) ] == '/' ) &&
			 (_strnicmp( this->name, fd.name, strlen( this->name) ) == 0 ) ) {
			return true;
		}
		//else
		return false;

	}

private:

} ;

class SERVICECONFIGDATA {

public:
	//Name of the service
	char*		scd_name;
	//DNS name of web site or IIS description of web site
	char*		scd_webServer;
    char*       scd_siteEntry;
	//If -1 then scd_webServer is the site's DNS name
	//If > 0 then scd_webServer is the site's IIS description and this value is
	//   the server instance associated with that description.
	DWORD		scd_ServerInstance;
    BOOL        scd_ignoreSuffixOn;

    char*   strFactors;

	SERVICECONFIGDATA::SERVICECONFIGDATA() {
		scd_name = NULL;
		scd_webServer = NULL;
        scd_siteEntry = NULL;
        scd_ServerInstance = -1;
        scd_ignoreSuffixOn = TRUE;
        strFactors = NULL;
        numFactors = 0;
        numIgnoreSuffixes = 0;
	}
	SERVICECONFIGDATA::~SERVICECONFIGDATA() {
		if ( scd_name != NULL ) free( scd_name );
        if ( scd_siteEntry != NULL ) free( scd_siteEntry );
		if ( scd_webServer != NULL ) free( scd_webServer );
        if ( strFactors != NULL ) free( strFactors );
	}
///UMBS requested workaround
	void addReqAuthNURI( BSTR	*bstr ) {
		FOLDERDATA	data( bstr, FALSE );
		svReqAuthN.push_back( data );
	}
	BOOL	isReqAuthN( char* szURI ) {
		FOLDERDATA	toFind( szURI );
		vector<FOLDERDATA>::iterator	iter;
		iter = find( svReqAuthN.begin(), svReqAuthN.end(), toFind );
		if ( iter != svReqAuthN.end() ) {
			return( TRUE );
		}
		return( FALSE );
	}
///end UMBS specific

	void addURI( BSTR*	bstr, int protectedStatus ) {
		FOLDERDATA	data( bstr, protectedStatus );
		svProtected.push_back( data );
	}

    void addFactor( BSTR* bstr ) {
        char*   factor = _com_util::ConvertBSTRToString( *bstr );
        FACTOR  f( _com_util::ConvertBSTRToString( *bstr ) );
        svFactors.push_back( f );

        if ( strFactors == NULL ) {
            strFactors = factor;
        } else {
            char*   newFactors;

            newFactors = (char*)malloc( sizeof( factor ) + sizeof( strFactors ) + 1 );
            sprintf( newFactors, "%s,%s", strFactors, factor );
            free( factor );
            strFactors = newFactors;
            //hot, strcat'ing akshun
        }
        numFactors++;
    }

    int getNumFactors () {
        return( numFactors );
    }

    BOOL removeSuffix( char* f ) {
        IGNORESUFFIX  toFind( f );
        vector<IGNORESUFFIX>::iterator iter;
        iter = find( svSuffixes.begin(), svSuffixes.end(), toFind );
        if ( iter != svSuffixes.end() ) {
            int pos = strlen( f ) - strlen( iter->suffix );
            (f)[ pos ] = '\0';
            return TRUE;
        }
        return FALSE;
    }

    // -2 Something bad happened
    // -1 Factor has a suffix that is being ignored
    //  0 Factor not found
    //  1 Factor found
    int getFactor( char **f ) {
        ///Be wary of static char []
        char    f2[ 512 ];
        BOOL    hasSuffix;
        strncpy( f2, *f, 512 );

        hasSuffix = removeSuffix( f2 );
        if ( hasSuffix && !scd_ignoreSuffixOn ) {
            return( -1 );
        }
        FACTOR  toFind( f2 );
        vector<FACTOR>::iterator iter;
        iter = find( svFactors.begin(), svFactors.end(), toFind );
        if ( iter != svFactors.end() ) {
            return( TRUE );
        }
        return( FALSE );
    }

    void addIgnoreSuffix( BSTR *bstr ) {
        IGNORESUFFIX    s( _com_util::ConvertBSTRToString( *bstr ) );
        svSuffixes.push_back( s );
        numIgnoreSuffixes++;
    }

	int protectedStatus( char* szURI ) {

        FOLDERDATA fd;
        int size, i, ps;
        size_t parentLength;
        /// Loop through the svProtected list
        /// If it satisfies the isChildCase() and the parent's strlen() is longer than the previous parent's strlen()
        /// then the protectedStatus is the same as the longer parent.

        parentLength = 0;
        ps = UNPROTECTED;
        size = svProtected.size();
        for ( i = 0; i < size; i++ ) {
            fd = svProtected.at( i );
            if ( ischildcase( (unsigned char*)szURI, (unsigned char*)fd.name, 0 ) ) {
                if ( strlen( fd.name ) > parentLength ) {
                    parentLength = strlen( fd.name );
                    ps = fd.protectedStatus;
                }
            }
        }        
        return( ps );
	}
	bool isEmpty() {
		return svProtected.empty();
	}
	int	numElements() {
		return (int)svProtected.size();
	}
	SERVICECONFIGDATA	& operator=( const SERVICECONFIGDATA &fd ) {
		if( this != &fd ) {
			this->scd_name = (char*)malloc( strlen( fd.scd_name ) + 1 );
			strcpy( this->scd_name, fd.scd_name );
			this->scd_webServer = (char*)malloc( strlen( fd.scd_webServer ) + 1 );
			strcpy( this->scd_webServer, fd.scd_webServer );
			this->svProtected = fd.svProtected;
		}
		return( *this );
	}

private:
	vector<FOLDERDATA> svProtected;
	vector<FOLDERDATA> svReqAuthN;
    vector<FACTOR>     svFactors;
    vector<IGNORESUFFIX>     svSuffixes;
    int numFactors;
    int numIgnoreSuffixes;	
};

class SERVICELIST {

public:
	SERVICELIST() {
	}

	vector<SERVICECONFIGDATA*>::iterator	findServerInstance( char* szServerName, DWORD dwInstanceID, char* szURI ) {
		vector<SERVICECONFIGDATA*>::iterator	iter;
		for ( iter = vscd.begin(); iter < vscd.end(); iter++ ) {
			if ( ((*iter)->scd_ServerInstance != -1 && (*iter)->scd_ServerInstance == dwInstanceID ) ||
				  ( strcmp( (*iter)->scd_webServer, szServerName ) == 0 ) ) {
					  if ( (*iter)->protectedStatus( szURI ) != UNPROTECTED ) {
						  return( iter );
					  }
			}
		}
		return( iter );

	}

	int	getProtectedStatus( char* szServerName, DWORD dwInstanceID, char* szURI, SERVICECONFIGDATA** scd ) {
		vector<SERVICECONFIGDATA*>::iterator	iter;
		int	status;

		for ( iter = vscd.begin(); iter < vscd.end(); iter++ ) {
			if ( ((*iter)->scd_ServerInstance != -1 && (*iter)->scd_ServerInstance == dwInstanceID ) ||
                ( (*iter)->scd_webServer != NULL && ( strcmp( (*iter)->scd_webServer, szServerName ) == 0 ) ) ) {
				if ( (status = (*iter)->protectedStatus( szURI )) != UNPROTECTED ) {
					*scd = (*iter);
					return( status );
				}
			}
		}
		return( UNPROTECTED );

	}
	void addService( SERVICECONFIGDATA* scd ) {
		vscd.push_back( scd );
	}
private:
	vector<SERVICECONFIGDATA*>	vscd;

};

int
LoadConfigFile( SERVICELIST* sl, char* szFileName, COSIGNSETTINGS* pcs );
DWORD
GetIISServerInstance( WCHAR*	webSiteName );
void
UnloadConfigFile( COSIGNSETTINGS* pcs );
void
LoadDefaultSettings( COSIGNSETTINGS* pcs );

