/*
 * Copyright (c) 2002 Regents of The University of Michigan.
 * All Rights Reserved.  See COPYRIGHT.
 */


#include <windows.h>
#include "argcargv.h"
#include <sys/types.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#include <openssl\ssl.h>

#include <openssl\err.h>

#include <libsnet\snet.h>
#include "sparse.h"

#include "log.h"
#include "CosignMsg.h"

#include "CosignFilterMain.h"
#include "CosignStructs.h"
#include "config.h"
#include "mkcookie.h"

static int connect_sn( struct connlist *, SSL_CTX *, int);
static int close_sn( SNET *);
void (*logger)( char * ) = NULL;
#define MIN(a,b)        ((a)<(b)?(a):(b))
struct timeval		timeout = { 10 * 60, 0 };
int cosign_protocol = 0;


    static int
netcheck_cookie( char *secant, struct sinfo *si, SNET *sn, SERVICECONFIGDATA* scd )
{
    int			ac;
    int         i;
    int         factorsUsed;
    char		*line;
    char		**av;
    struct timeval      tv;
    extern int		errno;

	CSDEBUG1( CS_INFO, "Netchecking this cookie: \"%s\"\n", secant );
    if ( snet_writeftv( sn, NULL, "CHECK %s\r\n", secant ) < 0 ) {
		CosignLog( CS_ERROR, "netcheck_cookie::snet_writeftv failed\n");
		return( COSIGN_ERROR );
    }

    tv = timeout;
    if (( line = snet_getline_multi( sn, logger, &tv )) == NULL ) {
		CosignLog( CS_ERROR, "netcheck_cookie::snet_getline_multi failed.\n" );
		return( COSIGN_ERROR );
    }

    switch( *line ) {
    case '2':
		CSDEBUG0( CS_INFO, "We like 200!\n" );
		break;

	case '4' : 
		CSDEBUG1( CS_INFO, "netcheck_cookie: %s\n", line );
		return( COSIGN_LOGGED_OUT );

    case '5':
		/* choose another connection */
		CSDEBUG1( CS_INFO, "Choose another connection: %s\n", line );
		return( COSIGN_RETRY );

    default:
		CosignLog( CS_ERROR, "netcheck_cookie: %s\n", line );
		return( COSIGN_ERROR );
    }

    if (( ac = argcargv( line, &av )) < 4 ) {
		CosignLog( CS_ERROR, "netcheck_cookie: wrong number of args: %s\n", line );
		return( COSIGN_ERROR );
    }

    /* I guess we check some sizing here :) */
	if ( strlen( av[ 1 ] ) >= sizeof( si->si_ipaddr ) ) {
		CosignLog( CS_ERROR,  "netcheck_cookie: IP address too long\n" );
		return( COSIGN_ERROR );
    }
    strcpy( si->si_ipaddr, av[ 1 ] );

	if ( strlen( av[ 2 ] ) >= sizeof( si->si_user ) ) {
		CosignLog( CS_ERROR, "netcheck_cookie: username too long\n" );
		return( COSIGN_ERROR );
    }
    strcpy( si->si_user, av[ 2 ] );

    //Grab the factors
    if ( cosign_protocol == 2 ) {
        CSDEBUG0( CS_INFO, "netcheck_cookie: retrieving factors\n" );
        factorsUsed = 0;
        for ( i = 3; i < ac; i++ ) {
            CSDEBUG1( CS_INFO, "Factor=%s\n", av[ i ] );
            switch ( scd->getFactor( &av[ i ] ) ) {
            case -2: // Something bad happened
                /// XXX REPORT ERROR TO LOGS!
                return( COSIGN_ERROR );
                break;
            case -1: // Required factor has a suffix that is being ignored
                CosignLog( CS_ERROR, "Factor %s matches with suffix, but suffix matching is off.\n", av[ i ] );
                return ( COSIGN_BAD_SUFFIX );
                break;
            case  0://  Required factor not found
                // Though the factor given was not required by the config, it will
                // still be added to the list of factors used.
                break;
            case  1://  Required factor found
                factorsUsed++;
                break;
            }
            if ( i != 3 ) {
                _snprintf( si->si_factor, sizeof( si->si_factor ), "%s %s", si->si_factor, av[ i ] );
            } else {
                _snprintf( si->si_factor, sizeof( si->si_factor ), "%s", av[ i ] );
            }
        }
        CSDEBUG1( CS_INFO, "si->si_factor = %s\n", si->si_factor );
        if ( factorsUsed != scd->getNumFactors() ) {
            ///What do we return here to signal that not enough factors were used?
            return( COSIGN_RETRY );
        }
    }

	if ( strlen( av[ 3 ] ) >= sizeof( si->si_realm ) ) {
		CosignLog( CS_ERROR, "netcheck_cookie: realm \"%s\" too long.\n", av[ 3 ] );
		return( COSIGN_ERROR );
    }
    strcpy( si->si_realm, av[ 3 ] );

    return( COSIGN_OK );
}

    static int
netretr_proxy( char *scookie, struct sinfo *si, SNET *sn, COSIGNSETTINGS *cs)
{
    HANDLE		fd;
    char		*line;
    char        path [ MAX_PATH ];
    char		buf[ 8192 ];
    DWORD       rr;
    struct timeval      tv;

    /* RETR service-cookie cookies */
    if ( snet_writeftv( sn, NULL, "RETR %s cookies\r\n", scookie ) < 0 ) {
	    CosignLog( CS_ERROR, "netretr_proxy: snet_writef failed");
	    return( COSIGN_ERROR );
    }
 
    /* name our file */
    if ( _snprintf( path, sizeof( path ), "%s\\%s", cs->cs_proxyDir, scookie ) >=
             sizeof( path )) {
	    CosignLog( CS_ERROR, "netretr_proxy: cookie path too long");
        return( COSIGN_ERROR );
    }
 
    /* open the file */
    fd = CreateFile( path, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
    if ( fd == INVALID_HANDLE_VALUE ) {
	    CosignLog( GetLastError(), "netretr_proxy::CreateFile( %s ) returned INVALID_HANDLE_VALUE\n", path );
	    return( COSIGN_ERROR );
    }
 
    tv = timeout;
    do {
	    if (( line = snet_getline( sn, &tv )) == NULL ) {
	        CosignLog( CS_ERROR, "netretr_proxy: snet_getline failed" );
	        CloseHandle( fd );
		    DeleteFile( path );
		    return( COSIGN_ERROR );
	    }	
 
	    switch( *line ) {
	    case '2':
	        break;
 
	    case '4':
	        CosignLog( CS_ERROR,  "netretr_proxy: %s", line );
		    CloseHandle( fd );
		    DeleteFile( path );
	        return( COSIGN_LOGGED_OUT );
 
	    case '5':
	        /* choose another connection */
	        CosignLog( CS_ERROR,  "netretr_proxy: 5xx" );
		    CloseHandle( fd );
		    DeleteFile( path );
	        return( COSIGN_RETRY );
 
	    default:
	        CosignLog( CS_ERROR,  "netretr_proxy: %s", line );
		    CloseHandle( fd );
		    DeleteFile( path );
		    return( COSIGN_ERROR );
	    }
 
	    if ( strlen( line ) < 3 ) {
	        CosignLog( CS_ERROR, "netretr_proxy: short line: %s", line );
		    CloseHandle( fd );
		    DeleteFile( path );
	        return( COSIGN_ERROR );
	    }
	    if ( !isdigit( (int)line[ 1 ] ) ||
	         !isdigit( (int)line[ 2 ] )) {
	        CosignLog( CS_ERROR, "netretr_proxy: bad response: %s", line );
		    CloseHandle( fd );
		    DeleteFile( path );
	        return( COSIGN_ERROR );
        }
 
	if ( line[ 3 ] != '\0' &&
		line[ 3 ] != ' ' &&
		line [ 3 ] != '-' ) {
	    CosignLog( CS_ERROR, "netretr_proxy: bad response: %s", line );
		CloseHandle( fd );
		DeleteFile( path );
	    return( COSIGN_ERROR );
	}

	if ( line[ 3 ] == '-' ) {
		if ( _snprintf( buf, sizeof( buf ), "x%s\n", &line[ 4 ] ) >= sizeof( buf )) {
			CosignLog( CS_ERROR, "netretr_proxy: data too long");
			CloseHandle( fd );
			DeleteFile( path );
			return( COSIGN_ERROR );
		}
		rr = strlen(buf);
		if ( WriteFile( fd, buf, rr, &rr, NULL ) == 0 ) {
			CosignLog( GetLastError(), "netretr_proxy::WriteFile(1) failed.\n");
			CloseHandle( fd );
			DeleteFile( path );
			return( COSIGN_ERROR );
		}
	}
 
} while ( line[ 3 ] == '-' );

if ( CloseHandle( fd ) == 0 ) {
	CosignLog( GetLastError(), "netretr_proxy::CloseHandle( %s ) failed.\n", path );
	DeleteFile( path );
	return( COSIGN_ERROR );
     }
 
     return( COSIGN_OK );
}

    static int
netretr_ticket( char *scookie, struct sinfo *si, SNET *sn, COSIGNSETTINGS* cs )
{
    char		*line, *kdir;
    char        tmpkrbfilename[ sizeof(si->si_krb5tktfile) ], krbpath [ MAX_PATH ];
    char		buf[ 8192 ];
    HANDLE		fd;
    size_t              size = 0;
    DWORD             rr;
    struct timeval      tv;
    extern int		errno;
    int         hashlen;

    /* clear it, in case we can't get it later */
    *si->si_krb5tktfile = '\0';

    /* RETR service-cookie TicketType */
    if ( snet_writeftv( sn, NULL, "RETR %s tgt\r\n", scookie ) < 0 ) {
	    CosignLog( CS_ERROR, "netretr_ticket: snet_writeftv failed.\n");
	    return( COSIGN_ERROR );
    }

    tv = timeout;
    if (( line = snet_getline_multi( sn, logger, &tv )) == NULL ) {
	    CosignLog( CS_ERROR, "netretr_ticket: snet_getline_multi: %s\n",
		    strerror( errno ));
	    return( COSIGN_ERROR );
    }

    switch( *line ) {
    case '2':
	    break;

    case '4':
	    CSDEBUG1( CS_INFO, "netretr_ticket: %s\n", line);
	    return( COSIGN_LOGGED_OUT );

    case '5':
	    /* choose another connection */
	    CSDEBUG1( CS_INFO, "choose another connection: %s\n", line );
	    return( COSIGN_RETRY );

    default:
	    CosignLog( CS_ERROR, "netcheck_cookie: %s\n", line );
	    return( COSIGN_ERROR );
    }

	hashlen = cs->cs_hashlen;
    if ( hashlen > sizeof( tmpkrbfilename )) {
        hashlen = sizeof( tmpkrbfilename );
    }
    if ( mkcookie( sizeof( tmpkrbfilename ), tmpkrbfilename ) != 0 ) {
		CosignLog( CS_ERROR, "mkcookie failed in netretr_ticket().\n" );
		return( COSIGN_ERROR );
    }
	if ( cs->cs_krbTicketDir == NULL || *(cs->cs_krbTicketDir) == '\0' ) {
		kdir = DEFAULT_TKT_PREFIX;
	} else {
		kdir = cs->cs_krbTicketDir;
	} 
	if ( _snprintf( krbpath, sizeof(krbpath), "%s\\%s", kdir, tmpkrbfilename ) > sizeof(krbpath)) {
		CosignLog( ERROR_INSUFFICIENT_BUFFER, "_snprintf failed in netretr_ticket().\n" );
		return( COSIGN_ERROR );
	}

    tv = timeout;
    if (( line = snet_getline( sn, &tv )) == NULL ) {
        CosignLog( CS_ERROR, "netretr_ticket for %s failed\n", scookie);
        return( COSIGN_ERROR );
    }
    size = atoi( line );

    fd = CreateFile( krbpath, GENERIC_WRITE, FILE_SHARE_READ, NULL,
		CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
	if ( fd == INVALID_HANDLE_VALUE ) {
		CosignLog( GetLastError(), "netretr_ticket::CreateFile( %s ) returned INVALID_HANDLE_VALUE\n", krbpath );
		return( COSIGN_ERROR );
    }

    /* Get file from server */
    while ( size > 0 ) {
        tv = timeout;
        if (( rr = snet_read( sn, buf, (int)MIN( sizeof( buf ), size ),
                &tv )) <= 0 ) {
            CosignLog( CS_ERROR, "retrieve tgt failed: %s\n", strerror( errno ));
            goto error2;
        }
        if ( WriteFile( fd, buf, rr, &rr, NULL ) == 0 ) {
			CosignLog( GetLastError(), "netretr_ticket::WriteFile(1) failed.\n");
            goto error2;
        }
        size -= rr;
    }
	if ( CloseHandle( fd ) == 0 ) {
		CosignLog( GetLastError(), "netretr_ticket::CloseHandle( %s ) failed.\n", krbpath );
        goto error1;
    }

    tv = timeout;
    if (( line = snet_getline( sn, &tv )) == NULL ) {
        CosignLog( CS_ERROR, "retr for %s failed: %s\n", scookie,
            strerror( errno ));
        goto error1;
    }
    if ( strcmp( line, "." ) != 0 ) {
        CSDEBUG1( CS_INFO, "%s", line );
        goto error1;
    }

    /* copy the path to the ticket file */
    strcpy( si->si_krb5tktfile, tmpkrbfilename );

    return( COSIGN_OK );

error2:
    CloseHandle( fd );
error1:
    DeleteFile( krbpath );
    return( COSIGN_ERROR );
}

    int
teardown_conn( struct connlist *cur )
{

    /* close down all children on exit */
    for ( ; cur != NULL; cur = cur->conn_next ) {
		if ( cur->conn_sn != NULL  ) {
			if ( close_sn( cur->conn_sn ) != 0 ) {
				CosignLog( CS_ERROR, "teardown_conn::close_sn failed\n" );
			}
		}
    }
    return( 0 );
}

int
CheckCookie( char *secant, struct sinfo *si, struct connlist **cl, SSL_CTX* ctx, int getkrbTicket, int getProxy, COSIGNSETTINGS *cs, SERVICECONFIGDATA* scd )
{
    struct connlist	**cur, *tmp;
    int			rc = COSIGN_ERROR;
	extern CRITICAL_SECTION	g_csConn;

    /* use connection, then shuffle if there is a problem
     * how do we reclaim once it's bad?
     * what happens if they are all bad?
     */

	EnterCriticalSection( &g_csConn );
    for ( cur = cl; *cur != NULL; cur = &(*cur)->conn_next ) {
		if ( (*cur)->conn_sn == NULL ) {
		    continue;
		}	
		CSDEBUG0( CS_INFO, "netcheck'ing_cookie\n" );
		if ( (rc = netcheck_cookie( secant, si, (*cur)->conn_sn, scd)) < 0 ) {
			if ( snet_close( (*cur)->conn_sn ) != 0 ) {
				fprintf( stderr, "choose_conn: snet_close failed\n" );
			}
			(*cur)->conn_sn = NULL;
		}
		if ( ( rc == COSIGN_LOGGED_OUT ) || ( rc == COSIGN_OK ) || (rc == COSIGN_BAD_SUFFIX) ) {
			goto done;
		}
    }

    /* all are closed or we didn't like their answer */
    for ( cur = cl; *cur != NULL; cur = &(*cur)->conn_next ) {
		if ( (*cur)->conn_sn != NULL ) {
		    continue;
		}
		CSDEBUG0( CS_INFO, "Attaching...\n" );
		if ( ctx == NULL ) {
			CSDEBUG0( CS_INFO, "CheckCookie, ctx is NULL.\n" );
		}
        if ( connect_sn( *cur, ctx, cs->cs_protocolVersionNeeded ) != 0 ) {
			CSDEBUG0( CS_INFO, "Attach failed!\n" );
			rc = COSIGN_ERROR;
			continue;
		}
		if ( (rc = netcheck_cookie( secant, si, (*cur)->conn_sn, scd)) < 0 ) {
			if ( snet_close( (*cur)->conn_sn ) != 0 ) {
				CSDEBUG0( CS_INFO, "choose_conn: snet_close failed.\n" );
			}
			(*cur)->conn_sn = NULL;
		}
		if ( ( rc == COSIGN_LOGGED_OUT ) || ( rc == COSIGN_OK )  || (rc == COSIGN_BAD_SUFFIX) ) {
			goto done;
		}
    }
	LeaveCriticalSection( &g_csConn );

	/* bug? we need to think about the mixed cases
	 * number of hosts we could actually talk to
	 * if we talk to anyone, we set and redirect
	 * otherwise, we are forbidden
	 */
	if ( rc == COSIGN_ERROR ) {
		return( COSIGN_ERROR );
	}
    return( COSIGN_RETRY );

done:
	CSDEBUG0( CS_INFO, "CheckCookie(), done:\n" );
    if ( cur != cl ) {
		tmp = *cur;
		*cur = (*cur)->conn_next;
		tmp->conn_next = *cl;
		*cl = tmp;
    }
	LeaveCriticalSection( &g_csConn );
    
    if ( rc == COSIGN_OK ) {
        if ( getkrbTicket && netretr_ticket( secant, si, (*cl)->conn_sn, cs) != 2 ) {
	        CosignLog( CS_ERROR, "Could not retrieve kerberos ticket.\n" );
	    }
	    if ( getProxy && netretr_proxy( secant, si, (*cl)->conn_sn, cs) != 2 ) {
			CosignLog( CS_ERROR, "Could not retrieve proxy cookie.\n" );
        }
	}
    CSDEBUG1( CS_INFO, "CheckCookie is returning %d\n", rc );
    return( rc );
}

    static int
connect_sn( struct connlist *cl, SSL_CTX *ctx, int protocolVersionNeeded )
{
    SOCKET			s;
    int         ac = 0;
    char*       line;
    char**      av;
#ifdef TLS
	char		buf[ 1024 ];
    X509		*peer;
	SSL_load_error_strings();
#endif
    struct timeval      tv;

    if (( s = socket( PF_INET, SOCK_STREAM, NULL )) == INVALID_SOCKET ) {
		CosignLog( WSAGetLastError(), "connect_sn::socket() failed.\n" );
	    return( -1 );
    }

    if ( connect( s, ( struct sockaddr *)&cl->conn_sin,
	    sizeof( struct sockaddr_in )) == SOCKET_ERROR ) {
		CosignLog( WSAGetLastError(), "connect_sn::connect() failed.\n" );
		(void)closesocket( s );
		return( -1 );
    }

	
    if (( cl->conn_sn = snet_attach( s, 1024 * 1024 ) ) == NULL ) {
		CosignLog( CS_ERROR, "connect_sn::snet_attach failed\n" );
		(void)closesocket( s );
		return( -1 );
    }

    tv = timeout;
    if (( line = snet_getline_multi( cl->conn_sn, logger, &tv )) == NULL ) {
		CosignLog( CS_ERROR,  "connect_sn::snet_getline_multi failed\n" );
		goto done;
    }
    if ( *line != '2' ) {
		CosignLog( CS_ERROR, "connect_sn: %s\n", line );
		goto done;
    }

    if ( (ac = argcargv( line, &av )) < 4 ) {
        CosignLog( CS_ERROR, "argcargv, %d.", ac );
        goto done;
    }
    if ( (cosign_protocol = atol( av[ 1 ] )) != 2 ) {
        if ( protocolVersionNeeded == 2 ) {
            ///CSReportEvent(); This needs a new cosignmsg.h thingy.
            CosignLog( CS_ERROR, "Required v2 protocol unsupported by server." );
            goto done;
        }
        ///This needs to go in the iiscosign log.
        CosignLog( CS_ERROR, "Falling back to protocol v0.\n" );
        cosign_protocol = 0;
    } else {
        cosign_protocol = 2;
    }

    if ( cosign_protocol == 2 ) {
        if ( snet_writeftv( cl->conn_sn, NULL, "STARTTLS 2\r\n" ) < 0 ) {
	    	CosignLog( CS_ERROR, "connect_sn: starttls 2 failed.\n" );
		    goto done;
        }
    } else {
        if ( snet_writeftv( cl->conn_sn, NULL, "STARTTLS\r\n" ) < 0 ) {
	    	CosignLog( CS_ERROR, "connect_sn: starttls failed.\n" );
		    goto done;
        }
    }
       
    tv = timeout;
    if (( line = snet_getline_multi( cl->conn_sn, logger, &tv )) == NULL ) {
		CosignLog( CS_ERROR, "connect_sn: snet_getline_multi failed\n" );
		goto done;
    }

    if ( *line != '2' ) {
		CosignLog( CS_ERROR, "connect_sn: %s\n", line );
		goto done;
    }

    if ( snet_starttls( cl->conn_sn, ctx, 0 ) != 1 ) {
		CosignLog( CS_ERROR, "connect_sn::snet_starttls failed.\n" );
		goto done;
    }

    if (( peer = SSL_get_peer_certificate( (SSL*)cl->conn_sn->sn_ssl )) == NULL ) {
		CosignLog( CS_ERROR, "connect_sn::SSL_get_peer_certificate, no certificate.\n" );
		goto done;
    }

    X509_NAME_get_text_by_NID( X509_get_subject_name( peer ), NID_commonName,
	    buf, sizeof( buf ));
    X509_free( peer );

    /// cn and host must match.  Need to add this check.
    ///    if ( strcasecmp( buf, cfg->host ) != 0 ) {
	///    CosignLog( CS_ERROR, "connect_sn: cn=%s & host=%s " 
	///	    "don't match!", buf, cfg->host );
	///    goto done;
    ///}
    
    if ( cosign_protocol == 2 ) {
        tv = timeout;
        if (( line = snet_getline_multi( cl->conn_sn, logger, &tv )) == NULL ) {
            CosignLog( CS_ERROR, "connect_sn::snet_getline_multi() failed" );
	        goto done;
	    }
	    if ( *line != '2' ) {
            CosignLog( CS_ERROR, "connect_sn:: starttls 2: %s", line );
	        goto done;
	    }
    }
	
    return( 0 );

done:
	ERR_free_strings();
    if ( snet_close( cl->conn_sn ) != 0 ) {
		CosignLog( CS_ERROR, "connect_sn::snet_close failed\n" );
    }
    return( -1 );
}


    static int
close_sn( SNET *sn )
{
    char		*line;
    struct timeval      tv;

    /* Close network connection */
    if ( snet_writeftv( sn, NULL, "QUIT\r\n", NULL ) < 0 ) {
		CosignLog( CS_ERROR, "close_sn::snet_writeftv failed\n" );
		return( -1 );
    }
    tv = timeout;
    if ( ( line = snet_getline_multi( sn, logger, &tv ) ) == NULL ) {
		CosignLog( CS_ERROR, "close_sn::snet_getline_multi failed\n" );
		return( -1 );
    }
    if ( *line != '2' ) {
		CosignLog( CS_ERROR, "close_sn: %s\n", line );
    }
    if ( snet_close( sn ) != 0 ) {
		CosignLog( CS_ERROR,  "close_sn::snet_close failed\n" );
		return( -1 );
    }
    return( 0 );
}

