
/*
 * Copyright (c) 2002, 2003 Regents of The University of Michigan.
 * All Rights Reserved.  See COPYRIGHT.
 */

#include <windows.h>
#include <stdarg.h>
#include <stdio.h>

#include <openssl\ssl.h>

#include <openssl\err.h>

#include <libsnet\snet.h>

#include "CosignMsg.h"
#include "CosignStructs.h"
#include "log.h"
#include "config.h"

#define ERRORBUFFERSIZE	2048
#define LOGFILENAME	"CosignLog.csl"
#define	LOGFILESUFFIX "csl"
#define	CSEVENTSOURCE	"Cosign"

char	g_szFullPath[ _MAX_PATH + 1 ];
char	g_szLogPath[ _MAX_PATH + 1 ];
int		g_UseEventViewer;
int		g_loggingMethod;
DWORD	g_dwMaxLogFileSize;
HANDLE	g_hLogFile = INVALID_HANDLE_VALUE;


void
CSReportEvent( WORD		wType,
			   WORD		wCategory,
			   DWORD	dwEventID,
			   WORD		wNumStrings,
			   LPCTSTR*	lpStrings )
{
	HANDLE hEventLog = RegisterEventSource( NULL, CSEVENTSOURCE );
	if ( hEventLog == NULL ) {
		///Should RegisterEventSource(kernel32.dll) to report missing Cosign event source.
		CosignLog( GetLastError(), "RegisterEventSource() failed.\n" );
	} else {
		CSDEBUG0( CS_INFO, "Reporting event...\n" );
		if ( ReportEvent( hEventLog,
			wType,
			wCategory,
			dwEventID,
			NULL,
			wNumStrings,
			0,
			lpStrings,
			NULL
			)  == 0 ){
			CosignLog( GetLastError(), "ReportEvent() failed.\n" );
		}	
		DeregisterEventSource( hEventLog );
	}
}
/*
 * InitDebugInfo
 * return values
 *	-1 -- Error setting up log file
 *   0 -- Everything went smoothly
 *   1 -- Already initialized
 */
	int
InitLog( int nMethod, int nLoggingMethod, char* szFilePath, DWORD dwMaxLogSize, char* szInitMsg )
{

	DWORD		dwBytesWritten;
	DWORD		dwErr;
	SYSTEMTIME	sysTime;
	char		szTime[ ERRORBUFFERSIZE ];

	if ( nMethod == 1 ) {
		//Use event viewer
		g_UseEventViewer = 1;
	} else {
		g_UseEventViewer = 0;
	}

	g_dwMaxLogFileSize = dwMaxLogSize;
	g_loggingMethod = nLoggingMethod;

	//If the log file already points to a valid handle, we assume that initialization has
	//already ocurred.  Of course, the person calling this function is smart and wouldn't
	//attempt to do this twice.  A little insurance never hurt, though.
	if ( g_hLogFile != INVALID_HANDLE_VALUE ) return( 1 );

	g_hLogFile = INVALID_HANDLE_VALUE;

	if ( strcmp( szFilePath, "" ) != 0 &&
		 ( strlen( szFilePath ) + strlen( LOGFILENAME ) < _MAX_PATH + 1 ) ) {
		strcpy( g_szLogPath, szFilePath );
		if ( (g_szLogPath[ strlen( g_szLogPath ) ] != '\\') ||
			 (g_szLogPath[ strlen( g_szLogPath ) ] != '/' ) ) {
			strcat( g_szLogPath, "\\" );
		}
		if ( _snprintf( g_szFullPath, sizeof( g_szFullPath ), "%s%s", g_szLogPath, LOGFILENAME ) > sizeof( g_szFullPath ) ) {
			///should probably realloc g_szFullPath since it is only allocated once
		}
		g_hLogFile = CreateFile( g_szFullPath, GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS | TRUNCATE_EXISTING,
			FILE_ATTRIBUTE_NORMAL, NULL );
		if ( g_hLogFile == INVALID_HANDLE_VALUE ) {
			g_hLogFile = CreateFile( g_szFullPath, GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS,
			FILE_ATTRIBUTE_NORMAL, NULL );
			if ( g_hLogFile  == INVALID_HANDLE_VALUE ) {
				OutputDebugString( "InitDebugInfo: Unable to create specified log file.\n" );
				/// CosignLog() this?
				dwErr = GetLastError();
				CosignLog( dwErr, "InitDebugInfo::CreateFile( %s ) failed!\n", g_szFullPath );
				if ( dwErr != ERROR_ACCESS_DENIED ) {
					return( -1 );
				}
			} else {
				//to make sure we skip the next if statement.
				dwErr = ERROR_SUCCESS;
			}
			if ( GetLastError() == ERROR_ACCESS_DENIED ) {
				CSDEBUG0( CS_INFO, "Creating a temporary file name...\n" );
				//Make up a new, unique, temporary file name.
				if ( GetTempFileName( g_szLogPath, LOGFILENAME, 0, g_szFullPath ) == 0 ) {
					CosignLog( GetLastError(), "InitDebugInfo::GetTempFileName( %s ) failed!\n", g_szLogPath );
					return( -1 );
				}
				g_hLogFile = CreateFile( g_szFullPath, GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS,
					FILE_ATTRIBUTE_NORMAL, NULL );
				if ( g_hLogFile == INVALID_HANDLE_VALUE ) {
					CosignLog( GetLastError(), "InitDebugInfo::CreateFile( %s ) failed!  Even tried making a unique file name!\n", g_szFullPath );
					return( -1 );
				}
			} 
		}
		GetLocalTime( &sysTime );
		//Don't care about buffer overflow.  Shouldn't happen, but if it does, will just be truncated.
		_snprintf( szTime, sizeof( szTime ), "%u-%u-%u\t%u:%u:%u\n",
			sysTime.wMonth,
			sysTime.wDay,
			sysTime.wYear,
			sysTime.wHour,
			sysTime.wMinute,
			sysTime.wSecond );
		if ( WriteFile( g_hLogFile, szTime, (DWORD)(strlen( szTime ) ), &dwBytesWritten, NULL ) == 0 ) {
			OutputDebugString( "InitDebugInfo::WriteFile( current time ) failed!\n" );
		}
		if ( szInitMsg != NULL && 
			WriteFile( g_hLogFile, szInitMsg, (DWORD)(strlen( szInitMsg ) ), &dwBytesWritten, NULL ) == 0 ) {
			OutputDebugString( "InitDebugInfo::WriteFile( szInitMsg ) failed!\n" );
		}

	}
	return( 0 );
}

	void
CloseLog()
{

	char	szLogFileName[ _MAX_PATH + 1 ];
	SYSTEMTIME	sysTime;

	if ( g_hLogFile == INVALID_HANDLE_VALUE ) {
		OutputDebugString( "CloseDebugInfo: Invalid handle to log file.\n" );
		return;
	}
	CloseHandle( g_hLogFile );
	g_hLogFile = INVALID_HANDLE_VALUE;

	GetLocalTime( &sysTime );
	//"MM-DD-YYYY-H-M-S.%s",
	//Don't care about buffer overflow.  Shouldn't happen, but if it does, will just be truncated.
	_snprintf( szLogFileName, sizeof( szLogFileName ), "%s%u-%u-%u-%u-%u-%u.%s",
		g_szLogPath,
		sysTime.wMonth,
		sysTime.wDay,
		sysTime.wYear,
		sysTime.wHour,
		sysTime.wMinute,
		sysTime.wSecond,
		LOGFILESUFFIX );
	OutputDebugString( szLogFileName );
	OutputDebugString( "\n" );
	OutputDebugString( g_szFullPath );
	OutputDebugString( "\n" );
	if ( CopyFile( g_szFullPath, szLogFileName, FALSE ) == 0 ) {
		OutputDebugString( "CloseDebugInfo::CopyFile() failed.\n" );
	}

}

	void
CosignLog( DWORD dwErrorCode, char* format, ... )
{

	LPVOID	err_msg;
	va_list	args;	
	DWORD	dwBufSize = ERRORBUFFERSIZE;
	DWORD	dwNumChars;
	char	szMsg[ ERRORBUFFERSIZE ] = "";
	HANDLE	hEventLog;
	LPCTSTR	strings[ 2 ];
	char*	lpSource = "CosignFilter";
	WORD	wNumStrings = 0;
	DWORD	dwBytesWritten;

	   
    va_start(args, format);
    dwNumChars = (DWORD) vsprintf( szMsg, format, args);
    va_end(args);
	wNumStrings++;
	strings[ 0 ] = szMsg;

    
	if ( dwNumChars >= dwBufSize ) {
		OutputDebugString( "Fatal Error!  CosignLog buffer is too small!\n" );
		return;
		///This is bad.
		///exit( sideways );
	}

	OutputDebugString( szMsg );

	/*
	 * Reporting major errors here.  So far that would be ERROR_ACCESS_DENIED.
	 */
	LPCTSTR	errAry[ 1 ] = { (char*)szMsg };
	switch( dwErrorCode ) {
		case ERROR_ACCESS_DENIED :
			CSReportEvent( EVENTLOG_ERROR_TYPE, 0, CS_ACCESSDENIED, 1, errAry );
			break;
		default:
			break;
	}

	if ( dwErrorCode < CS_ERROR_BASE ) {
		FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_ARGUMENT_ARRAY,
			NULL, dwErrorCode, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&err_msg,
			0, NULL );
		OutputDebugString( (LPTSTR)err_msg );
		wNumStrings++;
		strings[ 1 ] = (LPCTSTR)err_msg;
	}


	if ( g_UseEventViewer) {
		hEventLog = RegisterEventSource( NULL, lpSource );

		if ( !ReportEvent( hEventLog, EVENTLOG_INFORMATION_TYPE, (WORD)1, (DWORD)1,
			NULL, wNumStrings, 0, strings, NULL ) ) {
			//oops, oh no! Panic!
		}

		DeregisterEventSource( hEventLog );
	} 

	if ( g_hLogFile != INVALID_HANDLE_VALUE ) {
		
		SYSTEMTIME	sysTime;
		char		szTime[ ERRORBUFFERSIZE ];
		
		GetSystemTime( &sysTime );
		
		GetLocalTime( &sysTime );
		//Don't care about buffer overflow.  Shouldn't happen, but if it does, will just be truncated.
		_snprintf( szTime, sizeof( szTime), "[%u-%u-%u\t%u:%u:%u] ",
			sysTime.wMonth,
			sysTime.wDay,
			sysTime.wYear,
			sysTime.wHour,
			sysTime.wMinute,
			sysTime.wSecond );

		if ( WriteFile( g_hLogFile, szTime, (DWORD)(strlen( szTime ) ), &dwBytesWritten, NULL ) == 0 ) {
			OutputDebugString( "CosignLog::WriteFile( szTime ) failed!\n" );
		}
		if ( WriteFile( g_hLogFile, szMsg, (DWORD)(strlen( szMsg ) ), &dwBytesWritten, NULL ) == 0 ) {
			OutputDebugString( "CosignLog::WriteFile( szMsg ) failed!\n" );
		}

		if ( dwErrorCode < CS_ERROR_BASE && 
			WriteFile( g_hLogFile, err_msg, (DWORD)(strlen( (char *)err_msg ) ), &dwBytesWritten, NULL ) == 0 ) {
			OutputDebugString( "CosignLog::WriteFile( err_msg ) failed!\n" );
		}
		//Check file size
		LARGE_INTEGER	liSize;

		GetFileSizeEx( g_hLogFile, &liSize );

		if ( liSize.LowPart >= g_dwMaxLogFileSize ) {
			OutputDebugString( "Creating a new log file now.\n" );
			//close file
			CloseLog();
			InitLog( g_UseEventViewer, g_loggingMethod, g_szLogPath, g_dwMaxLogFileSize, "Previous log file surpassed size limit.  Creating new log file.\n\n" );
			//open a new one.
		}
	}
}
